# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_login import UserMixin, AnonymousUserMixin
from werkzeug.local import LocalProxy
import hashlib
import io
import os
import yaml
import uuid
import wrapt
import time

import logging

# noinspection PyCompatibility
from builtins import range, bytes

from octoprint.settings import settings as s

from octoprint.util import atomic_write, to_bytes, deprecated, monotonic_time, generate_api_key
from octoprint.util import get_fully_qualified_classname as fqcn

from octoprint.access.permissions import Permissions, OctoPrintPermission
from octoprint.access.groups import GroupChangeListener, Group

from past.builtins import basestring

class UserManager(GroupChangeListener, object):
	def __init__(self, group_manager, settings=None):
		self._group_manager = group_manager
		self._group_manager.register_listener(self)

		self._logger = logging.getLogger(__name__)
		self._session_users_by_session = dict()
		self._sessionids_by_userid = dict()
		self._enabled = True

		if settings is None:
			settings = s()
		self._settings = settings

		self._login_status_listeners = []

	def register_login_status_listener(self, listener):
		self._login_status_listeners.append(listener)

	def unregister_login_status_listener(self, listener):
		self._login_status_listeners.remove(listener)

	def anonymous_user_factory(self):
		if self.enabled:
			return AnonymousUser([self._group_manager.guest_group])
		else:
			return AdminUser([self._group_manager.admin_group, self._group_manager.user_group])

	def api_user_factory(self):
		return ApiUser([self._group_manager.admin_group, self._group_manager.user_group])

	@property
	def enabled(self):
		return self._enabled

	@enabled.setter
	def enabled(self, value):
		self._enabled = value

	def enable(self):
		self._enabled = True

	def disable(self):
		self._enabled = False

	def login_user(self, user):
		self._cleanup_sessions()

		if user is None or user.is_anonymous:
			return

		if isinstance(user, LocalProxy):
			# noinspection PyProtectedMember
			user = user._get_current_object()

		if not isinstance(user, User):
			return None

		if not isinstance(user, SessionUser):
			user = SessionUser(user)

		self._session_users_by_session[user.session] = user

		userid = user.get_id()
		if not userid in self._sessionids_by_userid:
			self._sessionids_by_userid[userid] = set()

		self._sessionids_by_userid[userid].add(user.session)

		for listener in self._login_status_listeners:
			try:
				listener.on_user_logged_in(user)
			except Exception:
				self._logger.exception("Error in on_user_logged_in on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

		self._logger.info("Logged in user: {}".format(user.get_id()))

		return user

	def logout_user(self, user, stale=False):
		if user is None or user.is_anonymous or isinstance(user, AdminUser):
			return

		if isinstance(user, LocalProxy):
			user = user._get_current_object()

		if not isinstance(user, SessionUser):
			return

		userid = user.get_id()
		sessionid = user.session

		if userid in self._sessionids_by_userid:
			try:
				self._sessionids_by_userid[userid].remove(sessionid)
			except KeyError:
				pass

		if sessionid in self._session_users_by_session:
			try:
				del self._session_users_by_session[sessionid]
			except KeyError:
				pass

		for listener in self._login_status_listeners:
			try:
				listener.on_user_logged_out(user, stale=stale)
			except Exception:
				self._logger.exception("Error in on_user_logged_out on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

		self._logger.info("Logged out user: {}".format(user.get_id()))

	def _cleanup_sessions(self):
		for session, user in list(self._session_users_by_session.items()):
			if not isinstance(user, SessionUser):
				continue
			if user.created + (24 * 60 * 60) < monotonic_time():
				self._logger.info("Cleaning up user session {} for user {}".format(session, user.get_id()))
				self.logout_user(user, stale=True)

	@staticmethod
	def create_password_hash(password, salt=None, settings=None):
		if not salt:
			if settings is None:
				settings = s()
			salt = settings.get(["accessControl", "salt"])
			if salt is None:
				import string
				from random import choice
				chars = string.ascii_lowercase + string.ascii_uppercase + string.digits
				salt = "".join(choice(chars) for _ in range(32))
				settings.set(["accessControl", "salt"], salt)
				settings.save()

		return hashlib.sha512(to_bytes(password, encoding="utf-8", errors="replace") + to_bytes(salt)).hexdigest()

	def check_password(self, username, password):
		user = self.find_user(username)
		if not user:
			return False

		hash = UserManager.create_password_hash(password,
		                                        settings=self._settings)
		if user.check_password(hash):
			# new hash matches, correct password
			return True
		else:
			# new hash doesn't match, but maybe the old one does, so check that!
			oldHash = UserManager.create_password_hash(password,
			                                           salt="mvBUTvwzBzD3yPwvnJ4E4tXNf3CGJvvW",
			                                           settings=self._settings)
			if user.check_password(oldHash):
				# old hash matches, we migrate the stored password hash to the new one and return True since it's the correct password
				self.change_user_password(username, password)
				return True
			else:
				# old hash doesn't match either, wrong password
				return False

	def add_user(self, username, password, active, permissions, groups, overwrite=False):
		pass

	def change_user_activation(self, username, active):
		pass

	def change_user_permissions(self, username, permissions):
		pass

	def add_permissions_to_user(self, username, permissions):
		pass

	def remove_permissions_from_user(self, username, permissions):
		pass

	def change_user_groups(self, username, groups):
		pass

	def add_groups_to_user(self, username, groups):
		pass

	def remove_groups_from_user(self, username, groups):
		pass

	def remove_groups_from_users(self, group):
		pass

	def change_user_password(self, username, password):
		pass

	def get_user_setting(self, username, key):
		return None

	def get_all_user_settings(self, username):
		return dict()

	def change_user_setting(self, username, key, value):
		pass

	def change_user_settings(self, username, new_settings):
		pass

	def remove_user(self, username):
		if username in self._sessionids_by_userid:
			sessions = self._sessionids_by_userid[username]
			for session in sessions:
				if session in self._session_users_by_session:
					del self._session_users_by_session[session]
			del self._sessionids_by_userid[username]

	def find_user(self, userid=None, session=None):
		if session is not None and session in self._session_users_by_session:
			user = self._session_users_by_session[session]
			if userid is None or userid == user.get_id():
				return user

		return None

	def find_sessions_for(self, matcher):
		result = []
		for user in self.get_all_users():
			if matcher(user):
				try:
					session_ids = self._sessionids_by_userid[user.get_id()]
					for session_id in session_ids:
						try:
							result.append(self._session_users_by_session[session_id])
						except KeyError:
							# unknown session after all
							continue
				except KeyError:
					# no session for user
					pass
		return result

	def get_all_users(self):
		return []

	def has_been_customized(self):
		return False

	def on_group_removed(self, group):
		self._logger.debug("Group {} got removed, removing from all users".format(group.key))
		self.remove_groups_from_users([group])

	def on_group_permissions_changed(self, group, added=None, removed=None):
		users = self.find_sessions_for(lambda u: group in u.groups)
		for listener in self._login_status_listeners:
			try:
				for user in users:
					listener.on_user_modified(user)
			except Exception:
				self._logger.exception("Error in on_user_modified on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

	def on_group_subgroups_changed(self, group, added=None, removed=None):
		users = self.find_sessions_for(lambda u: group in u.groups)
		for listener in self._login_status_listeners:
			# noinspection PyBroadException
			try:
				for user in users:
					listener.on_user_modified(user)
			except Exception:
				self._logger.exception("Error in on_user_modified on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

	def _trigger_on_user_modified(self, user):
		if isinstance(user, basestring):
			# user id
			users = []
			try:
				session_ids = self._sessionids_by_userid[user]
				for session_id in session_ids:
					try:
						users.append(self._session_users_by_session[session_id])
					except KeyError:
						# unknown session id
						continue
			except KeyError:
				# no session for user
				return
		elif isinstance(user, User) and not isinstance(user, SessionUser):
			users = self.find_sessions_for(lambda u: u.get_id() == user.get_id())
		elif isinstance(user, User):
			users = [user]
		else:
			return

		for listener in self._login_status_listeners:
			try:
				for user in users:
					listener.on_user_modified(user)
			except Exception:
				self._logger.exception("Error in on_user_modified on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

	def _trigger_on_user_removed(self, username):
		for listener in self._login_status_listeners:
			try:
				listener.on_user_removed(username)
			except Exception:
				self._logger.exception("Error in on_user_removed on {!r}".format(listener),
				                       extra=dict(callback=fqcn(listener)))

	#~~ Deprecated methods follow

	# TODO: Remove deprecated methods in OctoPrint 1.5.0

	@staticmethod
	def createPasswordHash(*args, **kwargs):
		"""
		.. deprecated: 1.4.0

		   Replaced by :func:`~UserManager.create_password_hash`
		"""
		# we can't use the deprecated decorator here since this method is static
		import warnings
		warnings.warn("createPasswordHash has been renamed to create_password_hash", DeprecationWarning, stacklevel=2)
		return UserManager.create_password_hash(*args, **kwargs)

	@deprecated("changeUserRoles has been replaced by change_user_permissions",
	            includedoc="Replaced by :func:`change_user_permissions`",
	            since="1.4.0")
	def changeUserRoles(self, username, roles):
		user = self.find_user(username)
		if user is None:
			raise UnknownUser(username)

		removed_roles = set(user._roles) - set(roles)
		self.removeRolesFromUser(username, removed_roles, user=user)

		added_roles = set(roles) - set(user._roles)
		self.addRolesToUser(username, added_roles, user=user)

	@deprecated("addRolesToUser has been replaced by add_permissions_to_user",
	            includedoc="Replaced by :func:`add_permissions_to_user`",
	            since="1.4.0")
	def addRolesToUser(self, username, roles, user=None):
		if user is None:
			user = self.find_user(username)

		if user is None:
			raise UnknownUser(username)

		if "admin" in roles:
			self.add_groups_to_user(username, self._group_manager.admin_group)

		if "user" in roles:
			self.remove_groups_from_user(username, self._group_manager.user_group)

	@deprecated("removeRolesFromUser has been replaced by remove_permissions_from_user",
	            includedoc="Replaced by :func:`remove_permissions_from_user`",
	            since="1.4.0")
	def removeRolesFromUser(self, username, roles, user=None):
		if user is None:
			user = self.find_user(username)

		if user is None:
			raise UnknownUser(username)

		if "admin" in roles:
			self.remove_groups_from_user(username, self._group_manager.admin_group)
			self.remove_permissions_from_user(username, Permissions.ADMIN)

		if "user" in roles:
			self.remove_groups_from_user(username, self._group_manager.user_group)

	checkPassword        = deprecated("checkPassword has been renamed to check_password",
	                                  includedoc="Replaced by :func:`check_password`",
	                                  since="1.4.0")(check_password)
	addUser              = deprecated("addUser has been renamed to add_user",
	                                  includedoc="Replaced by :func:`add_user`",
	                                  since="1.4.0")(add_user)
	changeUserActivation = deprecated("changeUserActivation has been renamed to change_user_activation",
	                                  includedoc="Replaced by :func:`change_user_activation`",
	                                  since="1.4.0")(change_user_activation)
	changeUserPassword   = deprecated("changeUserPassword has been renamed to change_user_password",
	                                  includedoc="Replaced by :func:`change_user_password`",
	                                  since="1.4.0")(change_user_password)
	getUserSetting       = deprecated("getUserSetting has been renamed to get_user_setting",
	                                  includedoc="Replaced by :func:`get_user_setting`",
	                                  since="1.4.0")(get_user_setting)
	getAllUserSettings   = deprecated("getAllUserSettings has been renamed to get_all_user_settings",
	                                  includedoc="Replaced by :func:`get_all_user_settings`",
	                                  since="1.4.0")(get_all_user_settings)
	changeUserSetting    = deprecated("changeUserSetting has been renamed to change_user_setting",
	                                  includedoc="Replaced by :func:`change_user_setting`",
	                                  since="1.4.0")(change_user_setting)
	changeUserSettings   = deprecated("changeUserSettings has been renamed to change_user_settings",
	                                  includedoc="Replaced by :func:`change_user_settings`",
	                                  since="1.4.0")(change_user_settings)
	removeUser           = deprecated("removeUser has been renamed to remove_user",
	                                  includedoc="Replaced by :func:`remove_user`",
	                                  since="1.4.0")(remove_user)
	findUser             = deprecated("findUser has been renamed to find_user",
	                                  includedoc="Replaced by :func:`find_user`",
	                                  since="1.4.0")(find_user)
	getAllUsers          = deprecated("getAllUsers has been renamed to get_all_users",
	                                  includedoc="Replaced by :func:`get_all_users`",
	                                  since="1.4.0")(get_all_users)
	hasBeenCustomized    = deprecated("hasBeenCustomized has been renamed to has_been_customized",
	                                  includedoc="Replaced by :func:`has_been_customized`",
	                                  since="1.4.0")(has_been_customized)


class LoginStatusListener(object):

	def on_user_logged_in(self, user):
		pass

	def on_user_logged_out(self, user, stale=False):
		pass

	def on_user_modified(self, user):
		pass

	def on_user_removed(self, userid):
		pass


##~~ FilebasedUserManager, takes available users from users.yaml file

class FilebasedUserManager(UserManager):
	def __init__(self, group_manager, path=None, settings=None):
		UserManager.__init__(self, group_manager, settings=settings)

		if path is None:
			path = self._settings.get(["accessControl", "userfile"])
			if path is None:
				path = os.path.join(s().getBaseFolder("base"), "users.yaml")

		self._userfile = path

		self._users = {}
		self._dirty = False

		self._customized = None
		self._load()

	def _load(self):
		if os.path.exists(self._userfile) and os.path.isfile(self._userfile):
			self._customized = True
			with io.open(self._userfile, 'rt', encoding='utf-8') as f:
				data = yaml.safe_load(f)
				for name, attributes in data.items():
					permissions = []
					if "permissions" in attributes:
						permissions = attributes["permissions"]

					groups = {self._group_manager.user_group}  # the user group is mandatory for all logged in users
					if "groups" in attributes:
						groups |= set(attributes["groups"])

					# migrate from roles to permissions
					if "roles" in attributes and not "permissions" in attributes:
						self._logger.info("Migrating user {} to new granular permission system".format(name))

						groups |= set(self._migrate_roles_to_groups(attributes["roles"]))
						self._dirty = True

					apikey = None
					if "apikey" in attributes:
						apikey = attributes["apikey"]
					settings = dict()
					if "settings" in attributes:
						settings = attributes["settings"]

					self._users[name] = User(username=name,
					                         passwordHash=attributes["password"],
					                         active=attributes["active"],
					                         permissions=self._to_permissions(*permissions),
					                         groups=self._to_groups(*groups),
					                         apikey=apikey,
					                         settings=settings)
					for sessionid in self._sessionids_by_userid.get(name, set()):
						if sessionid in self._session_users_by_session:
							self._session_users_by_session[sessionid].update_user(self._users[name])

			if self._dirty:
				self._save()

		else:
			self._customized = False

	def _save(self, force=False):
		if not self._dirty and not force:
			return

		data = {}
		for name, user in self._users.items():
			if not user or not isinstance(user, User):
				continue

			data[name] = {
				"password": user._passwordHash,
				"active": user._active,
				"groups": self._from_groups(*user._groups),
				"permissions": self._from_permissions(*user._permissions),
				"apikey": user._apikey,
				"settings": user._settings,

				# TODO: deprecated, remove in 1.5.0
				"roles": user._roles
			}

		with atomic_write(self._userfile, mode='wt', permissions=0o600, max_permissions=0o666) as f:
			yaml.safe_dump(data, f, default_flow_style=False, indent=4, allow_unicode=True)
			self._dirty = False
		self._load()

	def _migrate_roles_to_groups(self, roles):
		# If admin is inside the roles, just return admin group
		if "admin" in roles:
			return [self._group_manager.admin_group, self._group_manager.user_group]
		else:
			return [self._group_manager.user_group]

	def _refresh_groups(self, user):
		user._groups = self._to_groups(*map(lambda g: g.key, user.groups))

	def add_user(self, username, password, active=False, permissions=None, groups=None, apikey=None, overwrite=False):
		if not permissions:
			permissions = []
		permissions = self._to_permissions(*permissions)

		if not groups:
			groups = self._group_manager.default_groups
		groups = self._to_groups(*groups)

		if username in self._users and not overwrite:
			raise UserAlreadyExists(username)

		self._users[username] = User(username,
		                             UserManager.create_password_hash(password,
		                                                              settings=self._settings),
		                             active,
		                             permissions,
		                             groups,
		                             apikey=apikey)
		self._dirty = True
		self._save()

	def change_user_activation(self, username, active):
		if username not in self._users:
			raise UnknownUser(username)

		if self._users[username].is_active != active:
			self._users[username]._active = active
			self._dirty = True
			self._save()

			self._trigger_on_user_modified(username)

	def change_user_permissions(self, username, permissions):
		if username not in self._users:
			raise UnknownUser(username)

		user = self._users[username]

		permissions = self._to_permissions(*permissions)

		removed_permissions = list(set(user._permissions) - set(permissions))
		added_permissions = list(set(permissions) - set(user._permissions))

		if len(removed_permissions) > 0:
			user.remove_permissions_from_user(removed_permissions)
			self._dirty = True

		if len(added_permissions) > 0:
			user.add_permissions_to_user(added_permissions)
			self._dirty = True

		if self._dirty:
			self._save()
			self._trigger_on_user_modified(username)

	def add_permissions_to_user(self, username, permissions):
		if username not in self._users:
			raise UnknownUser(username)

		if self._users[username].add_permissions_to_user(self._to_permissions(*permissions)):
			self._dirty = True
			self._save()
			self._trigger_on_user_modified(username)

	def remove_permissions_from_user(self, username, permissions):
		if username not in self._users:
			raise UnknownUser(username)

		if self._users[username].remove_permissions_from_user(self._to_permissions(*permissions)):
			self._dirty = True
			self._save()
			self._trigger_on_user_modified(username)

	def remove_permissions_from_users(self, permissions):
		modified = []
		for user in self._users:
			dirty = user.remove_permissions_from_user(self._to_permissions(*permissions))
			if dirty:
				self._dirty = True
				modified.append(user.get_id())

		if self._dirty:
			self._save()
			for username in modified:
				self._trigger_on_user_modified(username)

	def change_user_groups(self, username, groups):
		if username not in self._users:
			raise UnknownUser(username)

		user = self._users[username]

		groups = self._to_groups(*groups)

		removed_groups = list(set(user._groups) - set(groups))
		added_groups = list(set(groups) - set(user._groups))

		if len(removed_groups):
			self._dirty |= user.remove_groups_from_user(removed_groups)
		if len(added_groups):
			self._dirty |= user.add_groups_to_user(added_groups)

		if self._dirty:
			self._save()
			self._trigger_on_user_modified(username)

	def add_groups_to_user(self, username, groups, save=True, notify=True):
		if username not in self._users:
			raise UnknownUser(username)

		if self._users[username].add_groups_to_user(self._to_groups(*groups)):
			self._dirty = True

			if save:
				self._save()

			if notify:
				self._trigger_on_user_modified(username)

	def remove_groups_from_user(self, username, groups, save=True, notify=True):
		if username not in self._users:
			raise UnknownUser(username)

		if self._users[username].remove_groups_from_user(self._to_groups(*groups)):
			self._dirty = True

			if save:
				self._save()

			if notify:
				self._trigger_on_user_modified(username)

	def remove_groups_from_users(self, groups):
		modified = []
		for username, user in self._users.items():
			dirty = user.remove_groups_from_user(self._to_groups(*groups))
			if dirty:
				self._dirty = True
				modified.append(username)

		if self._dirty:
			self._save()

			for username in modified:
				self._trigger_on_user_modified(username)

	def change_user_password(self, username, password):
		if not username in self._users:
			raise UnknownUser(username)

		passwordHash = UserManager.create_password_hash(password,
		                                                settings=self._settings)
		user = self._users[username]
		if user._passwordHash != passwordHash:
			user._passwordHash = passwordHash
			self._dirty = True
			self._save()

	def change_user_setting(self, username, key, value):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		old_value = user.get_setting(key)
		if not old_value or old_value != value:
			user.set_setting(key, value)
			self._dirty = self._dirty or old_value != value
			self._save()

	def change_user_settings(self, username, new_settings):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		for key, value in new_settings.items():
			old_value = user.get_setting(key)
			user.set_setting(key, value)
			self._dirty = self._dirty or old_value != value
		self._save()

	def get_all_user_settings(self, username):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		return user.get_all_settings()

	def get_user_setting(self, username, key):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		return user.get_setting(key)

	def generate_api_key(self, username):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		user._apikey = generate_api_key()
		self._dirty = True
		self._save()
		return user._apikey

	def delete_api_key(self, username):
		if not username in self._users:
			raise UnknownUser(username)

		user = self._users[username]
		user._apikey = None
		self._dirty = True
		self._save()

	def remove_user(self, username):
		UserManager.remove_user(self, username)

		if not username in self._users:
			raise UnknownUser(username)

		del self._users[username]
		self._dirty = True
		self._save()

	def find_user(self, userid=None, apikey=None, session=None):
		user = UserManager.find_user(self, userid=userid, session=session)

		if user is not None:
			return user

		if userid is not None:
			if userid not in self._users:
				return None
			return self._users[userid]

		elif apikey is not None:
			for user in self._users.values():
				if apikey == user._apikey:
					return user
			return None

		else:
			return None

	def get_all_users(self):
		return list(self._users.values())

	def has_been_customized(self):
		return self._customized

	def on_group_permissions_changed(self, group, added=None, removed=None):
		# refresh our group references
		for user in self.get_all_users():
			if group in user.groups:
				self._refresh_groups(user)

		# call parent
		UserManager.on_group_permissions_changed(self, group, added=added, removed=removed)

	def on_group_subgroups_changed(self, group, added=None, removed=None):
		# refresh our group references
		for user in self.get_all_users():
			if group in user.groups:
				self._refresh_groups(user)

		# call parent
		UserManager.on_group_subgroups_changed(self, group, added=added, removed=removed)

	#~~ Helpers

	def _to_groups(self, *groups):
		return list(set(filter(lambda x: x is not None,
		                       (self._group_manager._to_group(group) for group in groups))))

	def _to_permissions(self, *permissions):
		return list(set(filter(lambda x: x is not None,
		                       (Permissions.find(permission) for permission in permissions))))

	def _from_groups(self, *groups):
		return list(set(group.key for group in groups))

	def _from_permissions(self, *permissions):
		return list(set(permission.key for permission in permissions))

	# ~~ Deprecated methods follow

	# TODO: Remove deprecated methods in OctoPrint 1.5.0

	generateApiKey = deprecated("generateApiKey has been renamed to generate_api_key",
	                            includedoc="Replaced by :func:`generate_api_key`",
	                            since="1.4.0")(generate_api_key)
	deleteApiKey   = deprecated("deleteApiKey has been renamed to delete_api_key",
	                            includedoc="Replaced by :func:`delete_api_key`",
	                            since="1.4.0")(delete_api_key)
	addUser              = deprecated("addUser has been renamed to add_user",
	                                  includedoc="Replaced by :func:`add_user`",
	                                  since="1.4.0")(add_user)
	changeUserActivation = deprecated("changeUserActivation has been renamed to change_user_activation",
	                                  includedoc="Replaced by :func:`change_user_activation`",
	                                  since="1.4.0")(change_user_activation)
	changeUserPassword   = deprecated("changeUserPassword has been renamed to change_user_password",
	                                  includedoc="Replaced by :func:`change_user_password`",
	                                  since="1.4.0")(change_user_password)
	getUserSetting       = deprecated("getUserSetting has been renamed to get_user_setting",
	                                  includedoc="Replaced by :func:`get_user_setting`",
	                                  since="1.4.0")(get_user_setting)
	getAllUserSettings   = deprecated("getAllUserSettings has been renamed to get_all_user_settings",
	                                  includedoc="Replaced by :func:`get_all_user_settings`",
	                                  since="1.4.0")(get_all_user_settings)
	changeUserSetting    = deprecated("changeUserSetting has been renamed to change_user_setting",
	                                  includedoc="Replaced by :func:`change_user_setting`",
	                                  since="1.4.0")(change_user_setting)
	changeUserSettings   = deprecated("changeUserSettings has been renamed to change_user_settings",
	                                  includedoc="Replaced by :func:`change_user_settings`",
	                                  since="1.4.0")(change_user_settings)
	removeUser           = deprecated("removeUser has been renamed to remove_user",
	                                  includedoc="Replaced by :func:`remove_user`",
	                                  since="1.4.0")(remove_user)
	findUser             = deprecated("findUser has been renamed to find_user",
	                                  includedoc="Replaced by :func:`find_user`",
	                                  since="1.4.0")(find_user)
	getAllUsers          = deprecated("getAllUsers has been renamed to get_all_users",
	                                  includedoc="Replaced by :func:`get_all_users`",
	                                  since="1.4.0")(get_all_users)
	hasBeenCustomized    = deprecated("hasBeenCustomized has been renamed to has_been_customized",
	                                  includedoc="Replaced by :func:`has_been_customized`",
	                                  since="1.4.0")(has_been_customized)

##~~ Exceptions

class UserAlreadyExists(Exception):
	def __init__(self, username):
		Exception.__init__(self, "User %s already exists" % username)


class UnknownUser(Exception):
	def __init__(self, username):
		Exception.__init__(self, "Unknown user: %s" % username)


class UnknownRole(Exception):
	def _init_(self, role):
		Exception.__init__(self, "Unknown role: %s" % role)

##~~ Refactoring helpers

class MethodReplacedByBooleanProperty(object):

	def __init__(self, name, message, getter):
		self._name = name
		self._message = message
		self._getter = getter

	@property
	def _attr(self):
		return self._getter()

	def __call__(self):
		from warnings import warn
		warn(DeprecationWarning(self._message.format(name=self._name)), stacklevel=2)
		return self._attr

	def __eq__(self, other):
		return self._attr == other

	def __ne__(self, other):
		return self._attr != other

	def __bool__(self):
		# Python 3
		return self._attr

	def __nonzero__(self):
		# Python 2
		return self._attr

	def __hash__(self):
		return hash(self._attr)

	def __repr__(self):
		return "MethodReplacedByProperty({}, {}, {})".format(self._name, self._message, self._getter)

	def __str__(self):
		return str(self._attr)


# TODO: Remove compatibility layer in OctoPrint 1.5.0
class FlaskLoginMethodReplacedByBooleanProperty(MethodReplacedByBooleanProperty):

	def __init__(self, name, getter):
		message = "{name} is now a property in Flask-Login versions >= 0.3.0, which OctoPrint now uses. " + \
		          "Use {name} instead of {name}(). This compatibility layer will be removed in OctoPrint 1.5.0."
		MethodReplacedByBooleanProperty.__init__(self, name, message, getter)


# TODO: Remove compatibility layer in OctoPrint 1.5.0
class OctoPrintUserMethodReplacedByBooleanProperty(MethodReplacedByBooleanProperty):

	def __init__(self, name, getter):
		message = "{name} is now a property for consistency reasons with Flask-Login versions >= 0.3.0, which " + \
		          "OctoPrint now uses. Use {name} instead of {name}(). This compatibility layer will be removed " + \
		          "in OctoPrint 1.5.0."
		MethodReplacedByBooleanProperty.__init__(self, name, message, getter)

##~~ User object

class User(UserMixin):
	def __init__(self, username, passwordHash, active, permissions=None, groups=None, apikey=None, settings=None):
		if permissions is None:
			permissions = []
		if groups is None:
			groups = []

		self._username = username
		self._passwordHash = passwordHash
		self._active = active
		self._permissions = permissions
		self._groups = groups
		self._apikey = apikey

		if settings is None:
			settings = dict()

		self._settings = settings

	def as_dict(self):
		from octoprint.access.permissions import OctoPrintPermission
		return {
			"name": self._username,
			"active": bool(self.is_active),
			"permissions": list(map(lambda p: p.key, self._permissions)),
			"groups": list(map(lambda g: g.key, self._groups)),
			"needs": OctoPrintPermission.convert_needs_to_dict(self.needs),
			"apikey": self._apikey,
			"settings": self._settings,

			# TODO: deprecated, remove in 1.5.0
			"admin": self.has_permission(Permissions.ADMIN),
			"user": not self.is_anonymous,
			"roles": self._roles
		}

	def check_password(self, passwordHash):
		return self._passwordHash == passwordHash

	def get_id(self):
		return self.get_name()

	def get_name(self):
		return self._username

	@property
	def is_anonymous(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_anonymous", lambda: False)

	@property
	def is_authenticated(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_authenticated", lambda: True)

	@property
	def is_active(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_active", lambda: self._active)

	def get_all_settings(self):
		return self._settings

	def get_setting(self, key):
		if not isinstance(key, (tuple, list)):
			path = [key]
		else:
			path = key

		return self._get_setting(path)

	def set_setting(self, key, value):
		if not isinstance(key, (tuple, list)):
			path = [key]
		else:
			path = key
		return self._set_setting(path, value)

	def _get_setting(self, path):
		s = self._settings
		for p in path:
			if isinstance(s, dict) and p in s:
				s = s[p]
			else:
				return None
		return s

	def _set_setting(self, path, value):
		s = self._settings
		for p in path[:-1]:
			if p not in s:
				s[p] = dict()

			if not isinstance(s[p], dict):
				s[p] = dict()

			s = s[p]

		key = path[-1]
		s[key] = value
		return True

	def add_permissions_to_user(self, permissions):
		# Make sure the permissions variable is of type list
		if not isinstance(permissions, list):
			permissions = [permissions]

		assert(all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

		dirty = False
		for permission in permissions:
			if permissions not in self._permissions:
				self._permissions.append(permission)
				dirty = True

		return dirty

	def remove_permissions_from_user(self, permissions):
		# Make sure the permissions variable is of type list
		if not isinstance(permissions, list):
			permissions = [permissions]

		assert(all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

		dirty = False
		for permission in permissions:
			if permission in self._permissions:
				self._permissions.remove(permission)
				dirty = True

		return dirty

	def add_groups_to_user(self, groups):
		# Make sure the groups variable is of type list
		if not isinstance(groups, list):
			groups = [groups]

		assert(all(map(lambda p: isinstance(p, Group), groups)))

		dirty = False
		for group in groups:
			if group.is_toggleable() and group not in self._groups:
				self._groups.append(group)
				dirty = True

		return dirty

	def remove_groups_from_user(self, groups):
		# Make sure the groups variable is of type list
		if not isinstance(groups, list):
			groups = [groups]

		assert(all(map(lambda p: isinstance(p, Group), groups)))

		dirty = False
		for group in groups:
			if group.is_toggleable() and group in self._groups:
				self._groups.remove(group)
				dirty = True

		return dirty

	@property
	def permissions(self):
		if self._permissions is None:
			return []

		if Permissions.ADMIN in self._permissions:
			return Permissions.all()

		return list(filter(lambda p: p is not None, self._permissions))

	@property
	def groups(self):
		return list(self._groups)

	@property
	def effective_permissions(self):
		if self._permissions is None:
			return []
		return list(filter(lambda p: p is not None and self.has_permission(p), Permissions.all()))

	@property
	def needs(self):
		needs = set()

		for permission in self.permissions:
			if permission is not None:
				needs = needs.union(permission.needs)

		for group in self.groups:
			if group is not None:
				needs = needs.union(group.needs)

		return needs

	def has_permission(self, permission):
		return self.has_needs(*permission.needs)

	def has_needs(self, *needs):
		return set(needs).issubset(self.needs)

	def __repr__(self):
		return "User(id=%s,name=%s,active=%r,user=True,admin=%r,permissions=%s,groups=%s)" % (self.get_id(), self.get_name(), bool(self.is_active), self.has_permission(Permissions.ADMIN), self._permissions, self._groups)

	# ~~ Deprecated methods & properties follow

	# TODO: Remove deprecated methods & properties in OctoPrint 1.5.0

	asDict = deprecated("asDict has been renamed to as_dict",
	                    includedoc="Replaced by :func:`as_dict`",
	                    since="1.4.0")(as_dict)

	@property
	@deprecated("is_user is deprecated, please use has_permission", since="1.4.0")
	def is_user(self):
		return OctoPrintUserMethodReplacedByBooleanProperty("is_user", lambda: not self.is_anonymous)

	@property
	@deprecated("is_admin is deprecated, please use has_permission", since="1.4.0")
	def is_admin(self):
		return OctoPrintUserMethodReplacedByBooleanProperty("is_admin", lambda: self.has_permission(Permissions.ADMIN))

	@property
	@deprecated("roles is deprecated, please use has_permission", since="1.4.0")
	def roles(self):
		return self._roles

	@property
	def _roles(self):
		"""Helper for the deprecated self.roles and serializing to yaml"""
		if self.has_permission(Permissions.ADMIN):
			return ["user", "admin"]
		elif not self.is_anonymous:
			return ["user"]
		else:
			return []


class AnonymousUser(AnonymousUserMixin, User):
	def __init__(self, groups):
		User.__init__(self, None, "", True, [], groups)

	@property
	def is_anonymous(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_anonymous", lambda: True)

	@property
	def is_authenticated(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_authenticated", lambda: False)

	@property
	def is_active(self):
		return FlaskLoginMethodReplacedByBooleanProperty("is_active", lambda: self._active)

	def check_password(self, passwordHash):
		return True

	def as_dict(self):
		from octoprint.access.permissions import OctoPrintPermission
		return {
			"needs": OctoPrintPermission.convert_needs_to_dict(self.needs)
		}

	def __repr__(self):
		return "AnonymousUser(groups=%s)" % self._groups


class SessionUser(wrapt.ObjectProxy):
	def __init__(self, user):
		wrapt.ObjectProxy.__init__(self, user)

		self._self_session = "".join('%02X' % z for z in bytes(uuid.uuid4().bytes))
		self._self_created = monotonic_time()
		self._self_touched = monotonic_time()

	@property
	def session(self):
		return self._self_session

	@property
	def created(self):
		return self._self_created

	@property
	def touched(self):
		return self._self_touched

	def touch(self):
		self._self_touched = monotonic_time()

	@deprecated("SessionUser.get_session() has been deprecated, use SessionUser.session instead", since="1.3.5")
	def get_session(self):
		return self.session

	def update_user(self, user):
		self.__wrapped__ = user

	def as_dict(self):
		result = self.__wrapped__.as_dict()
		result.update(dict(session=self.session))
		return result

	def __repr__(self):
		return "SessionUser({!r},session={},created={})".format(self.__wrapped__, self.session, self.created)

##~~ User object to use when global api key is used to access the API

class ApiUser(User):
	def __init__(self, groups):
		User.__init__(self, "_api", "", True, [], groups)

##~~ User object to use when access control is disabled

class AdminUser(User):
	def __init__(self, groups):
		User.__init__(self, "_admin", "", True, [], groups)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os
import shutil
import time
import threading
import watchdog.events

import octoprint.filemanager
import octoprint.filemanager.util
import octoprint.util


class GcodeWatchdogHandler(watchdog.events.PatternMatchingEventHandler):

	"""
	Takes care of automatically "uploading" files that get added to the watched folder.
	"""

	def __init__(self, file_manager, printer):
		watchdog.events.PatternMatchingEventHandler.__init__(self, patterns=list(map(lambda x: "*.%s" % x, octoprint.filemanager.get_all_extensions())))

		self._logger = logging.getLogger(__name__)

		self._file_manager = file_manager
		self._printer = printer

	def initial_scan(self, folder):
		def run_scan():
			try:
				from os import scandir
			except ImportError:
				from scandir import scandir

			self._logger.info("Running initial scan on watched folder...")
			for entry in scandir(folder):
				path = entry.path

				if not self._valid_path(path):
					continue

				self._logger.info("Found {}, trying to add it".format(path))
				self._upload(path)
			self._logger.info("... initial scan done.")

		thread = threading.Thread(target=run_scan)
		thread.daemon = True
		thread.start()

	def on_created(self, event):
		path = event.src_path
		if not self._valid_path(path):
			return

		thread = threading.Thread(target=self._repeatedly_check, args=(path,))
		thread.daemon = True
		thread.start()

	def _upload(self, path):
		# noinspection PyBroadException
		try:
			file_wrapper = octoprint.filemanager.util.DiskFileWrapper(os.path.basename(path), path)

			# determine future filename of file to be uploaded, abort if it can't be uploaded
			try:
				futurePath, futureFilename = self._file_manager.sanitize(octoprint.filemanager.FileDestinations.LOCAL, file_wrapper.filename)
			except Exception:
				self._logger.exception("Could not wrap %s", path)
				futurePath = None
				futureFilename = None

			if futureFilename is None or (len(self._file_manager.registered_slicers) == 0 and not octoprint.filemanager.valid_file_type(futureFilename)):
				return

			# prohibit overwriting currently selected file while it's being printed
			futureFullPath = self._file_manager.join_path(octoprint.filemanager.FileDestinations.LOCAL, futurePath, futureFilename)
			futureFullPathInStorage = self._file_manager.path_in_storage(octoprint.filemanager.FileDestinations.LOCAL, futureFullPath)

			if not self._printer.can_modify_file(futureFullPathInStorage, False):
				return

			reselect = self._printer.is_current_file(futureFullPathInStorage, False)

			added_file = self._file_manager.add_file(octoprint.filemanager.FileDestinations.LOCAL,
			                                         file_wrapper.filename,
			                                         file_wrapper,
			                                         allow_overwrite=True)
			if os.path.exists(path):
				try:
					os.remove(path)
				except Exception:
					pass

			if reselect:
				self._printer.select_file(self._file_manager.path_on_disk(octoprint.filemanager.FileDestinations.LOCAL,
				                                                          added_file),
				                          False)
		except Exception:
			self._logger.exception("There was an error while processing the file {} in the watched folder".format(path))
		finally:
			if os.path.exists(path):
				# file is still there - that should only happen if something went wrong, so mark it as failed
				# noinspection PyBroadException
				try:
					shutil.move(path, "{}.failed".format(path))
				except:
					# something went really wrong here.... but we can't do anything about it, so just log it
					self._logger.exception("There was an error while trying to mark {} as failed in the watched folder".format(path))

	def _repeatedly_check(self, path, interval=1, stable=5):
		try:
			last_size = os.stat(path).st_size
		except Exception:
			return

		countdown = stable

		while True:
			try:
				new_size = os.stat(path).st_size
			except Exception:
				return

			if new_size == last_size:
				self._logger.debug("File at {} is no longer growing, counting down: {}".format(path, countdown))
				countdown -= 1
				if countdown <= 0:
					break
			else:
				self._logger.debug("File at {} is still growing (last: {}, new: {}), waiting...".format(path, last_size, new_size))
				countdown = stable

			last_size = new_size
			time.sleep(interval)

		self._logger.debug("File at {} is stable, moving it".format(path))
		self._upload(path)

	def _valid_path(self, path):
		_, ext = os.path.splitext(path)
		return octoprint.filemanager.valid_file_type(path) and not octoprint.util.is_hidden_path(path) and ext != "failed"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module contains printer profile related code.

A printer profile is a ``dict`` of the following structure:

.. list-table::
   :widths: 15 5 30
   :header-rows: 1

   * - Name
     - Type
     - Description
   * - ``id``
     - ``string``
     - Internal id of the printer profile
   * - ``name``
     - ``string``
     - Human readable name of the printer profile
   * - ``model``
     - ``string``
     - Printer model
   * - ``color``
     - ``string``
     - Color to associate with the printer profile
   * - ``volume``
     - ``dict``
     - Information about the print volume
   * - ``volume.width``
     - ``float``
     - Width of the print volume (X axis)
   * - ``volume.depth``
     - ``float``
     - Depth of the print volume (Y axis)
   * - ``volume.height``
     - ``float``
     - Height of the print volume (Z axis)
   * - ``volume.formFactor``
     - ``string``
     - Form factor of the print bed, either ``rectangular`` or ``circular``
   * - ``volume.origin``
     - ``string``
     - Location of gcode origin in the print volume, either ``lowerleft`` or ``center``
   * - ``volume.custom_box``
     - ``dict`` or ``False``
     - Custom boundary box overriding the default bounding box based on the provided width, depth, height and origin.
       If ``False``, the default boundary box will be used.
   * - ``volume.custom_box.x_min``
     - ``float``
     - Minimum valid X coordinate
   * - ``volume.custom_box.y_min``
     - ``float``
     - Minimum valid Y coordinate
   * - ``volume.custom_box.z_min``
     - ``float``
     - Minimum valid Z coordinate
   * - ``volume.custom_box.x_max``
     - ``float``
     - Maximum valid X coordinate
   * - ``volume.custom_box.y_max``
     - ``float``
     - Maximum valid Y coordinate
   * - ``volume.custom_box.z_max``
     - ``float``
     - Maximum valid Z coordinate
   * - ``heatedBed``
     - ``bool``
     - Whether the printer has a heated bed (``True``) or not (``False``)
   * - ``heatedChamber``
     - ``bool``
     - Whether the printer has a heated chamber (``True``) or not (``False``)
   * - ``extruder``
     - ``dict``
     - Information about the printer's extruders
   * - ``extruder.count``
     - ``int``
     - How many extruders the printer has (default 1)
   * - ``extruder.offsets``
     - ``list`` of ``tuple``
     - Extruder offsets relative to first extruder, list of (x, y) tuples, first is always (0,0)
   * - ``extruder.nozzleDiameter``
     - ``float``
     - Diameter of the printer nozzle(s)
   * - ``extruder.sharedNozzle``
     - ``boolean``
     - Whether there's only one nozzle shared among all extruders (true) or one nozzle per extruder (false).
   * - ``axes``
     - ``dict``
     - Information about the printer axes
   * - ``axes.x``
     - ``dict``
     - Information about the printer's X axis
   * - ``axes.x.speed``
     - ``float``
     - Speed of the X axis in mm/s
   * - ``axes.x.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed's origin (False, default) or towards it (True)
   * - ``axes.y``
     - ``dict``
     - Information about the printer's Y axis
   * - ``axes.y.speed``
     - ``float``
     - Speed of the Y axis in mm/s
   * - ``axes.y.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed's origin (False, default) or towards it (True)
   * - ``axes.z``
     - ``dict``
     - Information about the printer's Z axis
   * - ``axes.z.speed``
     - ``float``
     - Speed of the Z axis in mm/s
   * - ``axes.z.inverted``
     - ``bool``
     - Whether a positive value change moves the nozzle away from the print bed (False, default) or towards it (True)
   * - ``axes.e``
     - ``dict``
     - Information about the printer's E axis
   * - ``axes.e.speed``
     - ``float``
     - Speed of the E axis in mm/s
   * - ``axes.e.inverted``
     - ``bool``
     - Whether a positive value change extrudes (False, default) or retracts (True) filament

.. autoclass:: PrinterProfileManager
   :members:

.. autoclass:: BedFormFactor
   :members:

.. autoclass:: BedOrigin
   :members:

.. autoclass:: SaveError

.. autoclass:: CouldNotOverwriteError

.. autoclass:: InvalidProfileError
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import io
import os
import copy
import re
import logging

try:
	from os import scandir
except ImportError:
	from scandir import scandir

from octoprint.settings import settings
from octoprint.util import dict_merge, dict_sanitize, dict_contains_keys, is_hidden_path

class SaveError(Exception):
	"""Saving a profile failed"""
	pass

class CouldNotOverwriteError(SaveError):
	"""Overwriting of a profile not allowed"""
	pass

class InvalidProfileError(Exception):
	"""Profile invalid"""
	pass

class BedFormFactor(object):
	"""Valid values for bed form factor"""

	RECTANGULAR = "rectangular"
	"""Rectangular bed"""

	CIRCULAR = "circular"
	"""Circular bed"""

	@classmethod
	def values(cls):
		return [getattr(cls, name) for name in cls.__dict__ if not (name.startswith("__") or name == "values")]

BedTypes = BedFormFactor
"""Deprecated name of :class:`BedFormFactors`"""

class BedOrigin(object):
	"""Valid values for bed origin"""

	LOWERLEFT = "lowerleft"
	"""Origin at lower left corner of the bed when looking from the top"""

	CENTER = "center"
	"""Origin at the center of the bed when looking from top"""

	@classmethod
	def values(cls):
		return [getattr(cls, name) for name in cls.__dict__ if not (name.startswith("__") or name == "values")]

class PrinterProfileManager(object):
	"""
	Manager for printer profiles. Offers methods to select the globally used printer profile and to list, add, remove,
	load and save printer profiles.
	"""

	default = dict(
		id = "_default",
		name = "Default",
		model = "Generic RepRap Printer",
		color = "default",
		volume=dict(
			width = 200,
			depth = 200,
			height = 200,
			formFactor = BedFormFactor.RECTANGULAR,
			origin = BedOrigin.LOWERLEFT,
			custom_box = False
		),
		heatedBed = True,
		heatedChamber = False,
		extruder=dict(
			count = 1,
			offsets = [
				(0, 0)
			],
			nozzleDiameter = 0.4,
			sharedNozzle = False
		),
		axes=dict(
			x = dict(speed=6000, inverted=False),
			y = dict(speed=6000, inverted=False),
			z = dict(speed=200, inverted=False),
			e = dict(speed=300, inverted=False)
		)
	)

	def __init__(self):
		self._current = None
		self._folder = settings().getBaseFolder("printerProfiles")
		self._logger = logging.getLogger(__name__)

		self._migrate_old_default_profile()
		self._verify_default_available()

	def _migrate_old_default_profile(self):
		default_overrides = settings().get(["printerProfiles", "defaultProfile"])
		if not default_overrides:
			return

		if self.exists("_default"):
			return

		if not isinstance(default_overrides, dict):
			self._logger.warning("Existing default profile from settings is not a valid profile, refusing to migrate: {!r}".format(default_overrides))
			return

		default_overrides["id"] = "_default"
		self.save(default_overrides)

		settings().set(["printerProfiles", "defaultProfile"], None)
		settings().save()

		self._logger.info("Migrated default printer profile from settings to _default.profile: {!r}".format(default_overrides))

	def _verify_default_available(self):
		default_id = settings().get(["printerProfiles", "default"])
		if default_id is None:
			default_id = "_default"

		if not self.exists(default_id):
			if not self.exists("_default"):
				if default_id == "_default":
					self._logger.error("Profile _default does not exist, creating _default again and setting it as default")
				else:
					self._logger.error("Selected default profile {} and _default do not exist, creating _default again and setting it as default".format(default_id))
				self.save(self.__class__.default, allow_overwrite=True, make_default=True)
			else:
				self._logger.error("Selected default profile {} does not exists, resetting to _default".format(default_id))
				settings().set(["printerProfiles", "default"], "_default")
				settings().save()
			default_id = "_default"

		profile = self.get(default_id)
		if profile is None:
			self._logger.error("Selected default profile {} is invalid, resetting to default values".format(default_id))
			profile = copy.deepcopy(self.__class__.default)
			profile["id"] = default_id
			self.save(self.__class__.default, allow_overwrite=True, make_default=True)

	def select(self, identifier):
		if identifier is None or not self.exists(identifier):
			self._current = self.get_default()
			return False
		else:
			self._current = self.get(identifier)
			if self._current is None:
				self._logger.error("Profile {} is invalid, cannot select, falling back to default".format(identifier))
				self._current = self.get_default()
				return False
			else:
				return True

	def deselect(self):
		self._current = None

	def get_all(self):
		return self._load_all()

	def get(self, identifier):
		try:
			if self.exists(identifier):
				return self._load_from_path(self._get_profile_path(identifier))
			else:
				return None
		except InvalidProfileError:
			return None

	def remove(self, identifier, trigger_event=False):
		if self._current is not None and self._current["id"] == identifier:
			return False
		elif settings().get(["printerProfiles", "default"]) == identifier:
			return False

		removed = self._remove_from_path(self._get_profile_path(identifier))
		if removed and trigger_event:
			from octoprint.events import eventManager, Events
			payload = dict(identifier=identifier)
			eventManager().fire(Events.PRINTER_PROFILE_DELETED, payload=payload)

		return removed

	def save(self, profile, allow_overwrite=False, make_default=False, trigger_event=False):
		if "id" in profile:
			identifier = profile["id"]
		elif "name" in profile:
			identifier = profile["name"]
		else:
			raise InvalidProfileError("profile must contain either id or name")

		identifier = self._sanitize(identifier)
		profile["id"] = identifier

		self._migrate_profile(profile)
		profile = dict_sanitize(profile, self.__class__.default)
		profile = dict_merge(self.__class__.default, profile)

		path = self._get_profile_path(identifier)
		is_overwrite = os.path.exists(path)

		self._save_to_path(path, profile, allow_overwrite=allow_overwrite)

		if make_default:
			settings().set(["printerProfiles", "default"], identifier)
			settings().save()

		if self._current is not None and self._current["id"] == identifier:
			self.select(identifier)

		from octoprint.events import eventManager, Events
		if trigger_event:
			payload = dict(identifier=identifier)
			event = Events.PRINTER_PROFILE_MODIFIED if is_overwrite else Events.PRINTER_PROFILE_ADDED
			eventManager().fire(event, payload=payload)

		return self.get(identifier)

	def is_default_unmodified(self):
		default = settings().get(["printerProfiles", "default"])
		return default is None or default == "_default" or not self.exists("_default")

	@property
	def profile_count(self):
		return len(self._load_all_identifiers())

	@property
	def last_modified(self):
		dates = [os.stat(self._folder).st_mtime]
		dates += [entry.stat().st_mtime for entry in scandir(self._folder) if entry.name.endswith(".profile")]
		return max(dates)

	def get_default(self):
		default = settings().get(["printerProfiles", "default"])
		if default is not None and self.exists(default):
			profile = self.get(default)
			if profile is not None:
				return profile
			else:
				self._logger.warning("Default profile {} is invalid, falling back to built-in defaults".format(default))

		return copy.deepcopy(self.__class__.default)

	def set_default(self, identifier):
		all_identifiers = self._load_all_identifiers()
		if identifier is not None and not identifier in all_identifiers:
			return

		settings().set(["printerProfiles", "default"], identifier)
		settings().save()

	def get_current_or_default(self):
		if self._current is not None:
			return self._current
		else:
			return self.get_default()

	def get_current(self):
		return self._current

	def exists(self, identifier):
		if identifier is None:
			return False
		else:
			path = self._get_profile_path(identifier)
			return os.path.exists(path) and os.path.isfile(path)

	def _load_all(self):
		all_identifiers = self._load_all_identifiers()
		results = dict()
		for identifier, path in all_identifiers.items():
			try:
				profile = self._load_from_path(path)
			except InvalidProfileError:
				self._logger.warning("Profile {} is invalid, skipping".format(identifier))
				continue

			if profile is None:
				continue

			results[identifier] = dict_merge(self.__class__.default, profile)
		return results

	def _load_all_identifiers(self):
		results = dict()
		for entry in scandir(self._folder):
			if is_hidden_path(entry.name) or not entry.name.endswith(".profile"):
				continue

			if not entry.is_file():
				continue

			identifier = entry.name[:-len(".profile")]
			results[identifier] = entry.path
		return results

	def _load_from_path(self, path):
		if not os.path.exists(path) or not os.path.isfile(path):
			return None

		import yaml
		with io.open(path, 'rt', encoding='utf-8') as f:
			profile = yaml.safe_load(f)

		if profile is None or not isinstance(profile, dict):
			raise InvalidProfileError("Profile is None or not a dictionary")

		if self._migrate_profile(profile):
			try:
				self._save_to_path(path, profile, allow_overwrite=True)
			except Exception:
				self._logger.exception("Tried to save profile to {path} after migrating it while loading, ran into exception".format(path=path))

		profile = self._ensure_valid_profile(profile)

		if not profile:
			self._logger.warning("Invalid profile: %s" % path)
			raise InvalidProfileError()
		return profile

	def _save_to_path(self, path, profile, allow_overwrite=False):
		validated_profile = self._ensure_valid_profile(profile)
		if not validated_profile:
			raise InvalidProfileError()

		if os.path.exists(path) and not allow_overwrite:
			raise SaveError("Profile %s already exists and not allowed to overwrite" % profile["id"])

		from octoprint.util import atomic_write
		import yaml
		try:
			with atomic_write(path, mode='wt', max_permissions=0o666) as f:
				yaml.safe_dump(profile, f, default_flow_style=False, indent=2, allow_unicode=True)
		except Exception as e:
			self._logger.exception("Error while trying to save profile %s" % profile["id"])
			raise SaveError("Cannot save profile %s: %s" % (profile["id"], str(e)))

	def _remove_from_path(self, path):
		try:
			os.remove(path)
			return True
		except Exception:
			return False

	def _get_profile_path(self, identifier):
		return os.path.join(self._folder, "%s.profile" % identifier)

	def _sanitize(self, name):
		if name is None:
			return None

		if "/" in name or "\\" in name:
			raise ValueError("name must not contain / or \\")

		import string
		valid_chars = "-_.() {ascii}{digits}".format(ascii=string.ascii_letters, digits=string.digits)
		sanitized_name = ''.join(c for c in name if c in valid_chars)
		sanitized_name = sanitized_name.replace(" ", "_")
		return sanitized_name

	def _migrate_profile(self, profile):
		# make sure profile format is up to date
		modified = False

		if "volume" in profile and "formFactor" in profile["volume"] and not "origin" in profile["volume"]:
			profile["volume"]["origin"] = BedOrigin.CENTER if profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR else BedOrigin.LOWERLEFT
			modified = True

		if "volume" in profile and not "custom_box" in profile["volume"]:
			profile["volume"]["custom_box"] = False
			modified = True

		if "extruder" in profile and not "sharedNozzle" in profile["extruder"]:
			profile["extruder"]["sharedNozzle"] = False
			modified = True

		if "extruder" in profile and "sharedNozzle" in profile["extruder"] and profile["extruder"]["sharedNozzle"]:
			profile["extruder"]["offsets"] = [(0.0, 0.0)]
			modified = True

		if not "heatedChamber" in profile:
			profile["heatedChamber"] = False
			modified = True

		return modified

	def _ensure_valid_profile(self, profile):
		# ensure all keys are present
		if not dict_contains_keys(self.default, profile):
			self._logger.warning("Profile invalid, missing keys. Expected: {expected!r}. Actual: {actual!r}".format(expected=list(self.default.keys()), actual=list(profile.keys())))
			return False

		# conversion helper
		def convert_value(profile, path, converter):
			value = profile
			for part in path[:-1]:
				if not isinstance(value, dict) or not part in value:
					raise RuntimeError("%s is not contained in profile" % ".".join(path))
				value = value[part]

			if not isinstance(value, dict) or not path[-1] in value:
				raise RuntimeError("%s is not contained in profile" % ".".join(path))

			value[path[-1]] = converter(value[path[-1]])

		# convert ints
		for path in (("extruder", "count"), ("axes", "x", "speed"), ("axes", "y", "speed"), ("axes", "z", "speed")):
			try:
				convert_value(profile, path, int)
			except Exception as e:
				self._logger.warning("Profile has invalid value for path {path!r}: {msg}".format(path=".".join(path), msg=str(e)))
				return False

		# convert floats
		for path in (("volume", "width"), ("volume", "depth"), ("volume", "height"), ("extruder", "nozzleDiameter")):
			try:
				convert_value(profile, path, float)
			except Exception as e:
				self._logger.warning("Profile has invalid value for path {path!r}: {msg}".format(path=".".join(path), msg=str(e)))
				return False

		# convert booleans
		for path in (("axes", "x", "inverted"), ("axes", "y", "inverted"), ("axes", "z", "inverted"), ("extruder", "sharedNozzle")):
			try:
				convert_value(profile, path, bool)
			except Exception as e:
				self._logger.warning("Profile has invalid value for path {path!r}: {msg}".format(path=".".join(path), msg=str(e)))
				return False

		# validate form factor
		if not profile["volume"]["formFactor"] in BedFormFactor.values():
			self._logger.warning("Profile has invalid value volume.formFactor: {formFactor}".format(formFactor=profile["volume"]["formFactor"]))
			return False

		# validate origin type
		if not profile["volume"]["origin"] in BedOrigin.values():
			self._logger.warning("Profile has invalid value in volume.origin: {origin}".format(origin=profile["volume"]["origin"]))
			return False

		# ensure origin and form factor combination is legal
		if profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR and not profile["volume"]["origin"] == BedOrigin.CENTER:
			profile["volume"]["origin"] = BedOrigin.CENTER

		# force width and depth of volume to be identical for circular beds, with width being the reference
		if profile["volume"]["formFactor"] == BedFormFactor.CIRCULAR:
			profile["volume"]["depth"] = profile["volume"]["width"]

		# if we have a custom bounding box, validate it
		if profile["volume"]["custom_box"] and isinstance(profile["volume"]["custom_box"], dict):
			if not len(profile["volume"]["custom_box"]):
				profile["volume"]["custom_box"] = False

			else:
				default_box = self._default_box_for_volume(profile["volume"])
				for prop, limiter in (("x_min", min), ("y_min", min), ("z_min", min),
				                      ("x_max", max), ("y_max", max), ("z_max", max)):
					if prop not in profile["volume"]["custom_box"] or profile["volume"]["custom_box"][prop] is None:
						profile["volume"]["custom_box"][prop] = default_box[prop]
					else:
						value = profile["volume"]["custom_box"][prop]
						try:
							value = limiter(float(value), default_box[prop])
							profile["volume"]["custom_box"][prop] = value
						except Exception:
							self._logger.warning("Profile has invalid value in volume.custom_box.{}: {!r}".format(prop, value))
							return False

				# make sure we actually do have a custom box and not just the same values as the
				# default box
				for prop in profile["volume"]["custom_box"]:
					if profile["volume"]["custom_box"][prop] != default_box[prop]:
						break
				else:
					# exactly the same as the default box, remove custom box
					profile["volume"]["custom_box"] = False

		# validate offsets
		offsets = []
		for offset in profile["extruder"]["offsets"]:
			if not len(offset) == 2:
				self._logger.warning("Profile has an invalid extruder.offsets entry: {entry!r}".format(entry=offset))
				return False
			x_offset, y_offset = offset
			try:
				offsets.append((float(x_offset), float(y_offset)))
			except Exception:
				self._logger.warning("Profile has an extruder.offsets entry with non-float values: {entry!r}".format(entry=offset))
				return False
		profile["extruder"]["offsets"] = offsets

		return profile

	@staticmethod
	def _default_box_for_volume(volume):
		if volume["origin"] == BedOrigin.CENTER:
			half_width = volume["width"] / 2.0
			half_depth = volume["depth"] / 2.0
			return dict(x_min=-half_width,
			            x_max=half_width,
			            y_min=-half_depth,
			            y_max=half_depth,
			            z_min=0.0,
			            z_max=volume["height"])
		else:
			return dict(x_min=0.0,
			            x_max=volume["width"],
			            y_min=0.0,
			            y_max=volume["depth"],
			            z_min=0.0,
			            z_max=volume["height"])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

# Wrapper to the new access.users location

from octoprint.access.users import User, deprecated
from octoprint.access.users import *  # noqa: F403 ## possibly used by other modules

AccessUser = User

import warnings
warnings.warn("octoprint.users is deprecated, use octoprint.access.users instead", DeprecationWarning, stacklevel=2)

class User(AccessUser):
	@deprecated("octoprint.users.User is deprecated, please use octoprint.access.users.User instead")
	def __init__(self, username, passwordHash, active, roles, apikey=None, settings=None):
		from octoprint.server import groupManager

		if "admin" in roles:
			groups = [groupManager.admin_group]
		elif "user" in roles:
			groups = [groupManager.user_group]
		else:
			groups = [groupManager.guest_group]

		AccessUser(username=username, passwordHash=passwordHash, active=active, permissions=None, groups=groups,
		           apikey=apikey, settings=settings)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This "python package" doesn't actually install. This is intentional. It is merely
used to figure out some information about the environment a specific pip call
is running under (installation dir, whether it belongs to a virtual environment,
whether the install location is writable by the current user), and for that it
only needs to be invoked by pip, the pip call doesn't have to be successful
however.

If an environment variable "TESTBALLOON_OUTPUT" is set, it will be used as location
to write a file with the figured out data to. Simply writing to stdout (the default
behaviour if no such environment variable is set) is sadly not going to work out
with versions of pip > 8.0.0, which capture all stdout output regardless of used
--verbose or --log flags.
"""

import io
import os
import sys

def produce_output(stream):
	from distutils.command.install import install as cmd_install
	from distutils.dist import Distribution

	cmd = cmd_install(Distribution())
	cmd.finalize_options()

	install_dir = cmd.install_lib
	virtual_env = hasattr(sys, "real_prefix")
	writable = os.access(install_dir, os.W_OK)

	lines = ["PIP_INSTALL_DIR={}".format(install_dir),
	         "PIP_VIRTUAL_ENV={}".format(virtual_env),
	         "PIP_WRITABLE={}".format(writable)]

	for line in lines:
		print(line, file=stream)

	stream.flush()


path = os.environ.get("TESTBALLOON_OUTPUT", None)
if path is not None:
	# environment variable set, write to a log
	path = os.path.abspath(path)
	with io.open(path, 'wt+', encoding='utf-8') as output:
		produce_output(output)
else:
	# write to stdout
	produce_output(sys.stdout)

# fail intentionally
sys.exit(-1)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
click.disable_unicode_literals_warning = True
import logging

from octoprint.cli import pass_octoprint_ctx, OctoPrintContext, get_ctx_obj_option

#~~ "octoprint plugin:command" commands

class OctoPrintPluginCommands(click.MultiCommand):
	"""
	Custom `click.MultiCommand <http://click.pocoo.org/5/api/#click.MultiCommand>`_
	implementation that collects commands from the plugin hook
	:ref:`octoprint.cli.commands <sec-plugins-hook-cli-commands>`.

	.. attribute:: settings

	   The global :class:`~octoprint.settings.Settings` instance.

	.. attribute:: plugin_manager

	   The :class:`~octoprint.plugin.core.PluginManager` instance.
	"""

	sep = ":"

	def __init__(self, *args, **kwargs):
		click.MultiCommand.__init__(self, *args, **kwargs)

		self.settings = None
		self.plugin_manager = None
		self.hooks = dict()

		self._logger = logging.getLogger(__name__)
		self._initialized = False

	def _initialize(self, ctx):
		if self._initialized:
			return

		click.echo("Initializing settings & plugin subsystem...")
		if ctx.obj is None:
			ctx.obj = OctoPrintContext()

		# initialize settings and plugin manager based on provided
		# context (basedir and configfile)
		from octoprint import init_settings, init_pluginsystem, FatalStartupError
		try:
			self.settings = init_settings(get_ctx_obj_option(ctx, "basedir", None), get_ctx_obj_option(ctx, "configfile", None))
			self.plugin_manager = init_pluginsystem(self.settings,
			                                        safe_mode=get_ctx_obj_option(ctx, "safe_mode", False))
		except FatalStartupError as e:
			click.echo(e.message, err=True)
			click.echo("There was a fatal error initializing the settings or the plugin system.", err=True)
			ctx.exit(-1)

		# fetch registered hooks
		self.hooks = self.plugin_manager.get_hooks("octoprint.cli.commands")

		logging.basicConfig(level=logging.DEBUG if ctx.obj.verbosity > 0 else logging.WARN)

		self._initialized = True

	def list_commands(self, ctx):
		self._initialize(ctx)
		result = [name for name in self._get_commands()]
		result.sort()
		return result

	def get_command(self, ctx, cmd_name):
		self._initialize(ctx)
		commands = self._get_commands()
		return commands.get(cmd_name, None)

	def _get_commands(self):
		"""Fetch all commands from plugins providing any."""

		import collections
		result = collections.OrderedDict()

		for name, hook in self.hooks.items():
			try:
				commands = hook(self, pass_octoprint_ctx)
				for command in commands:
					if not isinstance(command, click.Command):
						self._logger.warning("Plugin {} provided invalid CLI command, ignoring it: {!r}".format(name, command))
						continue
					result[name + self.sep + command.name] = command
			except Exception:
				self._logger.exception("Error while retrieving cli commands for plugin {}".format(name),
				                       extra=dict(plugin=name))

		return result

@click.group()
@pass_octoprint_ctx
def plugin_commands(obj):
	pass

@plugin_commands.group(name="plugins", cls=OctoPrintPluginCommands)
def plugins():
	"""Additional commands provided by plugins."""
	pass




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
In this module the slicing support of OctoPrint is encapsulated.

.. autoclass:: SlicingProfile
   :members:

.. autoclass:: TemporaryProfile
   :members:

.. autoclass:: SlicingManager
   :members:
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import os

try:
	from os import scandir
except ImportError:
	from scandir import scandir

import octoprint.plugin
import octoprint.events
import octoprint.util
from octoprint.settings import settings

import logging

from .exceptions import UnknownSlicer, SlicerNotConfigured, SlicingCancelled, \
		ProfileAlreadyExists, ProfileException, CouldNotDeleteProfile, UnknownProfile


class SlicingProfile(object):
	"""
	A wrapper for slicing profiles, both meta data and actual profile data.

	Arguments:
	    slicer (str): Identifier of the slicer this profile belongs to.
	    name (str): Identifier of this slicing profile.
	    data (object): Profile data, actual structure depends on individual slicer implementation.
	    display_name (str): Displayable name for this slicing profile.
	    description (str): Description of this slicing profile.
	    default (bool): Whether this is the default slicing profile for the slicer.
	"""

	def __init__(self, slicer, name, data, display_name=None, description=None, default=False):
		self.slicer = slicer
		self.name = name
		self.data = data
		self.display_name = display_name
		self.description = description
		self.default = default


class TemporaryProfile(object):
	"""
	A wrapper for a temporary slicing profile to be used for a slicing job, based on a :class:`SlicingProfile` with
	optional ``overrides`` applied through the supplied ``save_profile`` method.

	Usage example:

	.. code-block:: python

	   temporary = TemporaryProfile(my_slicer.save_slicer_profile, my_default_profile,
	                                overrides=my_overrides)
	   with (temporary) as profile_path:
	       my_slicer.do_slice(..., profile_path=profile_path, ...)

	Arguments:
	    save_profile (callable): Method to use for saving the temporary profile, also responsible for applying the
	        supplied ``overrides``. This will be called according to the method signature of
	        :meth:`~octoprint.plugin.SlicerPlugin.save_slicer_profile`.
	    profile (SlicingProfile): The profile from which to derive the temporary profile.
	    overrides (dict): Optional overrides to apply to the ``profile`` for creation of the temporary profile.
	"""

	def __init__(self, save_profile, profile, overrides=None):
		self.save_profile = save_profile
		self.profile = profile
		self.overrides = overrides

	def __enter__(self):
		import tempfile
		temp_profile = tempfile.NamedTemporaryFile(prefix="slicing-profile-temp-", suffix=".profile", delete=False)
		temp_profile.close()

		self.temp_path = temp_profile.name
		self.save_profile(self.temp_path, self.profile, overrides=self.overrides)
		return self.temp_path

	def __exit__(self, type, value, traceback):
		import os
		try:
			os.remove(self.temp_path)
		except Exception:
			pass


class SlicingManager(object):
	"""
	The :class:`SlicingManager` is responsible for managing available slicers and slicing profiles.

	Arguments:
	    profile_path (str): Absolute path to the base folder where all slicing profiles are stored.
	    printer_profile_manager (~octoprint.printer.profile.PrinterProfileManager): :class:`~octoprint.printer.profile.PrinterProfileManager`
	       instance to use for accessing available printer profiles, most importantly the currently selected one.
	"""

	def __init__(self, profile_path, printer_profile_manager):
		self._logger = logging.getLogger(__name__)

		self._profile_path = profile_path
		self._printer_profile_manager = printer_profile_manager

		self._slicers = dict()
		self._slicer_names = dict()

	def initialize(self):
		"""
		Initializes the slicing manager by loading and initializing all available
		:class:`~octoprint.plugin.SlicerPlugin` implementations.
		"""
		self.reload_slicers()

	def reload_slicers(self):
		"""
		Retrieves all registered :class:`~octoprint.plugin.SlicerPlugin` implementations and registers them as
		available slicers.
		"""
		plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.SlicerPlugin)
		slicers = dict()
		for plugin in plugins:
			try:
				slicers[plugin.get_slicer_properties()["type"]] = plugin
			except Exception:
				self._logger.exception("Error while getting properties from slicer {}, ignoring it".format(plugin._identifier),
				                       extra=dict(plugin=plugin._identifier))
				continue
		self._slicers = slicers

	@property
	def slicing_enabled(self):
		"""
		Returns:
		    (boolean) True if there is at least one configured slicer available, False otherwise.
		"""
		return len(self.configured_slicers) > 0

	@property
	def registered_slicers(self):
		"""
		Returns:
		    (list of str) Identifiers of all available slicers.
		"""
		return list(self._slicers.keys())

	@property
	def configured_slicers(self):
		"""
		Returns:
		    (list of str) Identifiers of all available configured slicers.
		"""
		return list(map(lambda slicer: slicer.get_slicer_properties()["type"], filter(lambda slicer: slicer.is_slicer_configured(), self._slicers.values())))

	@property
	def default_slicer(self):
		"""
		Retrieves the default slicer.

		Returns:
		    (str) The identifier of the default slicer or ``None`` if the default slicer is not registered in the
		        system.
		"""
		slicer_name = settings().get(["slicing", "defaultSlicer"])
		if slicer_name in self.registered_slicers:
			return slicer_name
		else:
			return None

	def get_slicer(self, slicer, require_configured=True):
		"""
		Retrieves the slicer named ``slicer``. If ``require_configured`` is set to True (the default) an exception
		will be raised if the slicer is not yet configured.

		Arguments:
		    slicer (str): Identifier of the slicer to return
		    require_configured (boolean): Whether to raise an exception if the slicer has not been configured yet (True,
		        the default), or also return an unconfigured slicer (False).

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.SlicerNotConfigured: The ``slicer`` is not yet configured and ``require_configured`` was set to True.
		"""

		if not slicer in self._slicers:
			raise UnknownSlicer(slicer)

		if require_configured and not self._slicers[slicer].is_slicer_configured():
			raise SlicerNotConfigured(slicer)

		return self._slicers[slicer]

	def slice(self, slicer_name, source_path, dest_path, profile_name, callback,
	          callback_args=None, callback_kwargs=None, overrides=None,
	          on_progress=None, on_progress_args=None, on_progress_kwargs=None, printer_profile_id=None, position=None):
		"""
		Slices ``source_path`` to ``dest_path`` using slicer ``slicer_name`` and slicing profile ``profile_name``.
		Since slicing happens asynchronously, ``callback`` will be called when slicing has finished (either successfully
		or not), with ``callback_args`` and ``callback_kwargs`` supplied.

		If ``callback_args`` is left out, an empty argument list will be assumed for the callback. If ``callback_kwargs``
		is left out, likewise an empty keyword argument list will be assumed for the callback. Note that in any case
		the callback *must* support being called with the following optional keyword arguments:

		_analysis
		    If the slicer returned analysis data of the created machine code as part of its slicing result, this keyword
		    argument will contain that data.
		_error
		    If there was an error while slicing this keyword argument will contain the error message as returned from
		    the slicer.
		_cancelled
		    If the slicing job was cancelled this keyword argument will be set to True.

		Additionally callees may specify ``overrides`` for the specified slicing profile, e.g. a different extrusion
		temperature than defined in the profile or a different layer height.

		With ``on_progress``, ``on_progress_args`` and ``on_progress_kwargs``, callees may specify a callback plus
		arguments and keyword arguments to call upon progress reports from the slicing job. The progress callback will
		be called with a keyword argument ``_progress`` containing the current slicing progress as a value between 0
		and 1 plus all additionally specified args and kwargs.

		If a different printer profile than the currently selected one is to be used for slicing, its id can be provided
		via the keyword argument ``printer_profile_id``.

		If the ``source_path`` is to be a sliced at a different position than the print bed center, this ``position`` can
		be supplied as a dictionary defining the ``x`` and ``y`` coordinate in print bed coordinates of the model's center.

		Arguments:
		    slicer_name (str): The identifier of the slicer to use for slicing.
		    source_path (str): The absolute path to the source file to slice.
		    dest_path (str): The absolute path to the destination file to slice to.
		    profile_name (str): The name of the slicing profile to use.
		    callback (callable): A callback to call after slicing has finished.
		    callback_args (list or tuple): Arguments of the callback to call after slicing has finished. Defaults to
		        an empty list.
		    callback_kwargs (dict): Keyword arguments for the callback to call after slicing has finished, will be
		        extended by ``_analysis``, ``_error`` or ``_cancelled`` as described above! Defaults to an empty
		        dictionary.
		    overrides (dict): Overrides for the printer profile to apply.
		    on_progress (callable): Callback to call upon slicing progress.
		    on_progress_args (list or tuple): Arguments of the progress callback. Defaults to an empty list.
		    on_progress_kwargs (dict): Keyword arguments of the progress callback, will be extended by ``_progress``
		        as described above! Defaults to an empty dictionary.
		    printer_profile_id (str): Identifier of the printer profile for which to slice, if another than the
		        one currently selected is to be used.
		    position (dict): Dictionary containing the ``x`` and ``y`` coordinate in the print bed's coordinate system
		        of the sliced model's center. If not provided the model will be positioned at the print bed's center.
		        Example: ``dict(x=10,y=20)``.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer_name`` is unknown.
		    ~octoprint.slicing.exceptions.SlicerNotConfigured: The slice specified via ``slicer_name`` is not configured yet.
		"""

		if callback_args is None:
			callback_args = ()
		if callback_kwargs is None:
			callback_kwargs = dict()

		if not slicer_name in self.configured_slicers:
			if not slicer_name in self.registered_slicers:
				error = "No such slicer: {slicer_name}".format(**locals())
				exc = UnknownSlicer(slicer_name)
			else:
				error = "Slicer not configured: {slicer_name}".format(**locals())
				exc = SlicerNotConfigured(slicer_name)
			callback_kwargs.update(dict(_error=error, _exc=exc))
			callback(*callback_args, **callback_kwargs)
			raise exc

		slicer = self.get_slicer(slicer_name)

		printer_profile = None
		if printer_profile_id is not None:
			printer_profile = self._printer_profile_manager.get(printer_profile_id)

		if printer_profile is None:
			printer_profile = self._printer_profile_manager.get_current_or_default()

		def slicer_worker(slicer, model_path, machinecode_path, profile_name, overrides, printer_profile, position, callback, callback_args, callback_kwargs):
			try:
				slicer_name = slicer.get_slicer_properties()["type"]
				with self._temporary_profile(slicer_name, name=profile_name, overrides=overrides) as profile_path:
					ok, result = slicer.do_slice(
						model_path,
						printer_profile,
						machinecode_path=machinecode_path,
						profile_path=profile_path,
						position=position,
						on_progress=on_progress,
						on_progress_args=on_progress_args,
						on_progress_kwargs=on_progress_kwargs
					)

				if not ok:
					callback_kwargs.update(dict(_error=result))
				elif result is not None and isinstance(result, dict) and "analysis" in result:
					callback_kwargs.update(dict(_analysis=result["analysis"]))
			except SlicingCancelled:
				callback_kwargs.update(dict(_cancelled=True))
			finally:
				callback(*callback_args, **callback_kwargs)

		import threading
		slicer_worker_thread = threading.Thread(target=slicer_worker,
		                                        args=(slicer, source_path, dest_path, profile_name, overrides, printer_profile, position, callback, callback_args, callback_kwargs))
		slicer_worker_thread.daemon = True
		slicer_worker_thread.start()

	def cancel_slicing(self, slicer_name, source_path, dest_path):
		"""
		Cancels the slicing job on slicer ``slicer_name`` from ``source_path`` to ``dest_path``.

		Arguments:
		    slicer_name (str): Identifier of the slicer on which to cancel the job.
		    source_path (str): The absolute path to the source file being sliced.
		    dest_path (str): The absolute path to the destination file being sliced to.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer_name`` is unknown.
		"""

		slicer = self.get_slicer(slicer_name)
		slicer.cancel_slicing(dest_path)

	def load_profile(self, slicer, name, require_configured=True):
		"""
		Loads the slicing profile for ``slicer`` with the given profile ``name`` and returns it. If it can't be loaded
		due to an :class:`IOError` ``None`` will be returned instead.

		If ``require_configured`` is True (the default) a :class:`SlicerNotConfigured` exception will be raised
		if the indicated ``slicer`` has not yet been configured.

		Returns:
		    SlicingProfile: The requested slicing profile or None if it could not be loaded.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer specified via ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer specified via ``slicer`` has not yet been configured and
		        ``require_configured`` was True.
		    ~octoprint.slicing.exceptions.UnknownProfile: The profile for slicer ``slicer`` named ``name`` does not exist.
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		try:
			path = self.get_profile_path(slicer, name, must_exist=True)
		except IOError:
			return None
		return self._load_profile_from_path(slicer, path, require_configured=require_configured)

	def save_profile(self, slicer, name, profile, overrides=None, allow_overwrite=True, display_name=None, description=None):
		"""
		Saves the slicer profile ``profile`` for slicer ``slicer`` under name ``name``.

		``profile`` may be either a :class:`SlicingProfile` or a :class:`dict`.

		If it's a :class:`SlicingProfile`, its :attr:`~SlicingProfile.slicer``, :attr:`~SlicingProfile.name` and - if
		provided - :attr:`~SlicingProfile.display_name` and :attr:`~SlicingProfile.description` attributes will be
		overwritten with the supplied values.

		If it's a :class:`dict`, a new :class:`SlicingProfile` instance will be created with the supplied meta data and
		the profile data as the :attr:`~SlicingProfile.data` attribute.

		.. note::

		   If the profile is the first profile to be saved for the slicer, it will automatically be marked as default.

		Arguments:
		    slicer (str): Identifier of the slicer for which to save the ``profile``.
		    name (str): Identifier under which to save the ``profile``.
		    profile (SlicingProfile or dict): The :class:`SlicingProfile` or a :class:`dict` containing the profile
		        data of the profile the save.
		    overrides (dict): Overrides to apply to the ``profile`` before saving it.
		    allow_overwrite (boolean): If True (default) if a profile for the same ``slicer`` of the same ``name``
		        already exists, it will be overwritten. Otherwise an exception will be thrown.
		    display_name (str): The name to display to the user for the profile.
		    description (str): A description of the profile.

		Returns:
		    SlicingProfile: The saved profile (including the applied overrides).

		Raises:
		    ValueError: The supplied ``profile`` is neither a :class:`SlicingProfile` nor a :class:`dict`.
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.ProfileAlreadyExists: A profile with name ``name`` already exists for ``slicer`` and ``allow_overwrite`` is
		        False.
		"""
		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		if not isinstance(profile, SlicingProfile):
			if isinstance(profile, dict):
				profile = SlicingProfile(slicer, name, profile, display_name=display_name, description=description)
			else:
				raise ValueError("profile must be a SlicingProfile or a dict")
		else:
			profile.slicer = slicer
			profile.name = name
			if display_name is not None:
				profile.display_name = display_name
			if description is not None:
				profile.description = description

		first_profile = len(self.all_profiles(slicer, require_configured=False)) == 0

		path = self.get_profile_path(slicer, name)
		is_overwrite = os.path.exists(path)

		if is_overwrite and not allow_overwrite:
			raise ProfileAlreadyExists(slicer, profile.name)

		self._save_profile_to_path(slicer, path, profile, overrides=overrides, allow_overwrite=allow_overwrite)

		payload = dict(slicer=slicer,
		               profile=name)
		event = octoprint.events.Events.SLICING_PROFILE_MODIFIED if is_overwrite else octoprint.events.Events.SLICING_PROFILE_ADDED
		octoprint.events.eventManager().fire(event, payload)

		if first_profile:
			# enforce the first profile we add for this slicer  is set as default
			self.set_default_profile(slicer, name)

		return profile

	def _temporary_profile(self, slicer, name=None, overrides=None):
		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		profile = self._get_default_profile(slicer)
		if name:
			try:
				profile = self.load_profile(slicer, name)
			except (UnknownProfile, IOError):
				# in that case we'll use the default profile
				pass

		return TemporaryProfile(self.get_slicer(slicer).save_slicer_profile, profile, overrides=overrides)

	def delete_profile(self, slicer, name):
		"""
		Deletes the profile ``name`` for the specified ``slicer``.

		If the profile does not exist, nothing will happen.

		Arguments:
		    slicer (str): Identifier of the slicer for which to delete the profile.
		    name (str): Identifier of the profile to delete.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.CouldNotDeleteProfile: There was an error while deleting the profile.
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		if not name:
			raise ValueError("name must be set")

		try:
			try:
				path = self.get_profile_path(slicer, name, must_exist=True)
			except UnknownProfile:
				return
			os.remove(path)
		except ProfileException as e:
			raise e
		except Exception as e:
			raise CouldNotDeleteProfile(slicer, name, cause=e)
		else:
			octoprint.events.eventManager().fire(octoprint.events.Events.SLICING_PROFILE_DELETED, dict(slicer=slicer, profile=name))

	def set_default_profile(self, slicer, name, require_configured=False,
	                        require_exists=True):
		"""
		Sets the given profile as default profile for the slicer.

		Arguments:
		    slicer (str): Identifier of the slicer for which to set the default
		        profile.
		    name (str): Identifier of the profile to set as default.
		    require_configured (bool): Whether the slicer needs to be configured
		        for the action to succeed. Defaults to false. Will raise a
		        SlicerNotConfigured error if true and the slicer has not been
		        configured yet.
		    require_exists (bool): Whether the profile is required to exist in
		        order to be set as default. Defaults to true. Will raise a
		        UnknownProfile error if true and the profile is unknown.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer``
		        is unknown
		    ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer ``slicer``
		        has not yet been configured and ``require_configured`` was true.
		    ~octoprint.slicing.exceptions.UnknownProfile: The profile ``name``
		        was unknown for slicer ``slicer`` and ``require_exists`` was
		        true.
		"""
		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)
		if require_configured and not slicer in self.configured_slicers:
			raise SlicerNotConfigured(slicer)

		if not name:
			raise ValueError("name must be set")

		if require_exists and not name in self.all_profiles(slicer, require_configured=require_configured):
			raise UnknownProfile(slicer, name)

		default_profiles = settings().get(["slicing", "defaultProfiles"])
		if not default_profiles:
			default_profiles = dict()
		default_profiles[slicer] = name
		settings().set(["slicing", "defaultProfiles"], default_profiles)
		settings().save(force=True)

	def all_profiles(self, slicer, require_configured=False):
		"""
		Retrieves all profiles for slicer ``slicer``.

		If ``require_configured`` is set to True (default is False), only will return the profiles if the ``slicer``
		is already configured, otherwise a :class:`SlicerNotConfigured` exception will be raised.

		Arguments:
		    slicer (str): Identifier of the slicer for which to retrieve all slicer profiles
		    require_configured (boolean): Whether to require the slicer ``slicer`` to be already configured (True)
		        or not (False, default). If False and the slicer is not yet configured, a :class:`~octoprint.slicing.exceptions.SlicerNotConfigured`
		        exception will be raised.

		Returns:
		    dict of SlicingProfile: A dict of all :class:`SlicingProfile` instances available for the slicer ``slicer``, mapped by the identifier.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.SlicerNotConfigured: The slicer ``slicer`` is not configured and ``require_configured`` was True.
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)
		if require_configured and not slicer in self.configured_slicers:
			raise SlicerNotConfigured(slicer)

		slicer_profile_path = self.get_slicer_profile_path(slicer)
		return self.get_slicer(slicer, require_configured=False).get_slicer_profiles(slicer_profile_path)

	def profiles_last_modified(self, slicer):
		"""
		Retrieves the last modification date of ``slicer``'s profiles.

		Args:
		    slicer (str): the slicer for which to retrieve the last modification date

		Returns:
		    (float) the time stamp of the last modification of the slicer's profiles
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		slicer_profile_path = self.get_slicer_profile_path(slicer)
		return self.get_slicer(slicer, require_configured=False).get_slicer_profiles_lastmodified(slicer_profile_path)

	def get_slicer_profile_path(self, slicer):
		"""
		Retrieves the path where the profiles for slicer ``slicer`` are stored.

		Arguments:
		    slicer (str): Identifier of the slicer for which to retrieve the path.

		Returns:
		    str: The absolute path to the folder where the slicer's profiles are stored.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		path = os.path.join(self._profile_path, slicer)
		if not os.path.exists(path):
			os.makedirs(path)
		return path

	def get_profile_path(self, slicer, name, must_exist=False):
		"""
		Retrieves the path to the profile named ``name`` for slicer ``slicer``.

		If ``must_exist`` is set to True (defaults to False) a :class:`UnknownProfile` exception will be raised if the
		profile doesn't exist yet.

		Arguments:
		    slicer (str): Identifier of the slicer to which the profile belongs to.
		    name (str): Identifier of the profile for which to retrieve the path.
		    must_exist (boolean): Whether the path must exist (True) or not (False, default).

		Returns:
		    str: The absolute path to the profile identified by ``name`` for slicer ``slicer``.

		Raises:
		    ~octoprint.slicing.exceptions.UnknownSlicer: The slicer ``slicer`` is unknown.
		    ~octoprint.slicing.exceptions.UnknownProfile: The profile named ``name`` doesn't exist and ``must_exist`` was True.
		"""

		if not slicer in self.registered_slicers:
			raise UnknownSlicer(slicer)

		if not name:
			raise ValueError("name must be set")

		name = self._sanitize(name)

		path = os.path.join(self.get_slicer_profile_path(slicer), "{name}.profile".format(name=name))
		if not os.path.realpath(path).startswith(os.path.realpath(self._profile_path)):
			raise IOError("Path to profile {name} tried to break out of allows sub path".format(**locals()))
		if must_exist and not (os.path.exists(path) and os.path.isfile(path)):
			raise UnknownProfile(slicer, name)
		return path

	def _sanitize(self, name):
		if name is None:
			return None

		if "/" in name or "\\" in name:
			raise ValueError("name must not contain / or \\")

		import string
		valid_chars = "-_.() {ascii}{digits}".format(ascii=string.ascii_letters, digits=string.digits)
		sanitized_name = ''.join(c for c in name if c in valid_chars)
		sanitized_name = sanitized_name.replace(" ", "_")
		return sanitized_name

	def _load_profile_from_path(self, slicer, path, require_configured=False):
		profile = self.get_slicer(slicer, require_configured=require_configured).get_slicer_profile(path)
		default_profiles = settings().get(["slicing", "defaultProfiles"])
		if default_profiles and slicer in default_profiles:
			profile.default = default_profiles[slicer] == profile.name
		return profile

	def _save_profile_to_path(self, slicer, path, profile, allow_overwrite=True, overrides=None, require_configured=False):
		self.get_slicer(slicer, require_configured=require_configured).save_slicer_profile(path, profile, allow_overwrite=allow_overwrite, overrides=overrides)

	def _get_default_profile(self, slicer):
		default_profiles = settings().get(["slicing", "defaultProfiles"])
		if default_profiles and slicer in default_profiles:
			try:
				return self.load_profile(slicer, default_profiles[slicer])
			except (UnknownProfile, IOError):
				# in that case we'll use the slicers predefined default profile
				pass

		return self.get_slicer(slicer).get_slicer_default_profile()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import requests
import logging
import base64

BRANCH_HEAD_URL = "https://api.bitbucket.org/2.0/repositories/{user}/{repo}/commit/{branch}"

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.bitbucket_commit")


def _get_latest_commit(user, repo, branch, api_user=None, api_password=None):
	from ..exceptions import NetworkError

	url = BRANCH_HEAD_URL.format(user=user, repo=repo, branch=branch)
	headers = {}
	if api_user is not None and api_password is not None:
		auth_value = base64.b64encode(b"{user}:{pw}".format(user=api_user, pw=api_password))
		headers["authorization"] = "Basic {}".format(auth_value)

	try:
		r = requests.get(url, headers=headers, timeout=(3.05, 30))
	except requests.ConnectionError as exc:
		raise NetworkError(cause=exc)

	if not r.status_code == requests.codes.ok:
		return None

	reference = r.json()
	if not "hash" in reference:
		return None

	return reference["hash"]


def get_latest(target, check, online=True):
	from ..exceptions import ConfigurationInvalid

	if "user" not in check or "repo" not in check:
		raise ConfigurationInvalid("Update configuration for %s of type bitbucket_commit needs all of user and repo" % target)

	branch = "master"
	if "branch" in check and check["branch"] is not None:
		branch = check["branch"]

	api_user = check.get("api_user")
	api_password = check.get("api_password")

	current = check.get("current")

	information = dict(
		local=dict(name="Commit {commit}".format(commit=current if current is not None else "unknown"), value=current),
		remote=dict(name="?", value="?"),
		needs_online=not check.get("offline", False)
	)
	if not online and information["needs_online"]:
		return information, True

	remote_commit = _get_latest_commit(check["user"], check["repo"], branch, api_user, api_password)
	remote_name = "Commit {commit}".format(commit=remote_commit) if remote_commit is not None else "-"

	information["remote"] = dict(name=remote_name, value=remote_commit)
	is_current = (current is not None and current == remote_commit) or remote_commit is None

	logger.debug("Target: %s, local: %s, remote: %s" % (target, current, remote_commit))

	return information, is_current




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


class NoUpdateAvailable(Exception):
	pass

class UpdateAlreadyInProgress(Exception):
	pass

class UnknownUpdateType(Exception):
	pass

class UnknownCheckType(Exception):
	pass

class NetworkError(Exception):
	def __init__(self, message=None, cause=None):
		Exception.__init__(self)
		self.message = message
		self.cause = cause

	def __str__(self):
		if self.message is not None:
			return self.message
		elif self.cause is not None:
			return "NetworkError caused by {}".format(self.cause)
		else:
			return "NetworkError"

class UpdateError(Exception):
	def __init__(self, message, data):
		self.message = message
		self.data = data

class ScriptError(Exception):
	def __init__(self, returncode, stdout, stderr):
		self.returncode = returncode
		self.stdout = stdout
		self.stderr = stderr

class RestartFailed(Exception):
	pass

class ConfigurationInvalid(Exception):
	pass

class CannotCheckOffline(Exception):
	pass

class CannotUpdateOffline(Exception):
	pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from octoprint.server import user_permission
from octoprint.access import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.events import Events

import flask
from flask_babel import gettext

class Prompt(object):

	def __init__(self, text):
		self.text = text
		self.choices = []

		self._active = False

	@property
	def active(self):
		return self._active

	def add_choice(self, text):
		self.choices.append(text)

	def activate(self):
		self._active = True

	def validate_choice(self, choice):
		return 0 <= choice < len(self.choices)


class ActionCommandPromptPlugin(octoprint.plugin.AssetPlugin,
                                octoprint.plugin.EventHandlerPlugin,
                                octoprint.plugin.SettingsPlugin,
                                octoprint.plugin.SimpleApiPlugin,
                                octoprint.plugin.TemplatePlugin):

	COMMAND = "M876"
	CAP_PROMPT_SUPPORT = "PROMPT_SUPPORT"

	# noinspection PyMissingConstructor
	def __init__(self):
		self._prompt = None
		self._enable = "detected"
		self._command = None
		self._enable_emergency_sending = False
		self._enable_signal_support = False
		self._cap_prompt_support = False

	def initialize(self):
		self._enable = self._settings.get(["enable"])
		self._command = self._settings.get(["command"])
		self._enable_emergency_sending = self._settings.get_boolean(["enable_emergency_sending"])
		self._enable_signal_support = self._settings.get_boolean(["enable_signal_support"])

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="INTERACT",
			     name="Interact with printer prompts",
			     description=gettext("Allows to see and interact with printer prompts"),
			     default_groups=[USER_GROUP],
			     roles=["interact"])
		]

	#~ AssetPlugin

	def get_assets(self):
		return dict(js=["js/action_command_prompt.js"],
		            clientjs=["clientjs/action_command_prompt.js"])

	#~ EventHandlerPlugin

	def on_event(self, event, payload):
		if event == Events.CONNECTED and self._enable == "always" and self._enable_signal_support:
			self._printer.commands(["{command} P1".format(command=self._command)])
		elif event == Events.DISCONNECTED:
			self._close_prompt()

	#~ SettingsPlugin

	def get_settings_defaults(self):
		return dict(enable="detected",
		            command=self.COMMAND,
		            enable_emergency_sending=True,
		            enable_signal_support=True)

	def on_settings_save(self, data):
		octoprint.plugin.SettingsPlugin.on_settings_save(self, data)
		self._enable = self._settings.get(["enable"])
		self._command = self._settings.get(["command"])
		self._enable_emergency_sending = self._settings.get_boolean(["enable_emergency_sending"])
		self._enable_signal_support = self._settings.get_boolean(["enable_signal_support"])

	#~ SimpleApiPlugin

	def get_api_commands(self):
		return dict(select=["choice"])

	def on_api_command(self, command, data):
		if command == "select":
			if not Permissions.PLUGIN_ACTION_COMMAND_PROMPT_INTERACT.can():
				return flask.abort(403, "Insufficient permissions")

			if self._prompt is None:
				return flask.abort(409, "No active prompt")

			choice = data["choice"]
			if not isinstance(choice, int) or not self._prompt.validate_choice(choice):
				return flask.abort(400, "{!r} is not a valid value for choice".format(choice))

			self._answer_prompt(choice)

	def on_api_get(self, request):
		if not Permissions.PLUGIN_ACTION_COMMAND_PROMPT_INTERACT.can():
			return flask.abort(403, "Insufficient permissions")
		if self._prompt is None:
			return flask.jsonify()
		else:
			return flask.jsonify(text=self._prompt.text, choices=self._prompt.choices)

	#~ TemplatePlugin

	def get_template_configs(self):
		return [dict(type="settings", name=gettext("Action Command Prompt"), custom_bindings=False)]

	#~ action command handler

	def action_command_handler(self, comm, line, action, *args, **kwargs):
		if not action.startswith("prompt_"):
			return

		parts = action.split(None, 1)
		if len(parts) == 1:
			action = parts[0]
			parameter = ""
		else:
			action, parameter = parts

		if action == "prompt_begin":
			if self._prompt is not None and self._prompt.active:
				self._logger.warning("Prompt is already defined")
				return
			self._prompt = Prompt(parameter.strip())

		elif action == "prompt_choice" or action == "prompt_button":
			if self._prompt is None:
				return
			if self._prompt.active:
				self._logger.warning("Prompt is already active")
				return
			self._prompt.add_choice(parameter.strip())

		elif action == "prompt_show":
			if self._prompt is None:
				return
			if self._prompt.active:
				self._logger.warning("Prompt is already active")
				return
			self._show_prompt()

		elif action == "prompt_end":
			if self._prompt is None:
				return
			self._close_prompt()
			self._prompt = None

	#~ queuing handling

	def gcode_queuing_handler(self, comm_instance, phase, cmd, cmd_type, gcode, subcode=None, tags=None, *args, **kwargs):
		if gcode != self._command:
			return

		if self._enable == "never" or (self._enable == "detected" and not self._cap_prompt_support) or not self._enable_emergency_sending:
			return

		if not "S" in cmd:
			# we only force-send M876 Sx
			return

		# noinspection PyProtectedMember
		return comm_instance._emergency_force_send(cmd, "Force-sending {} to the printer".format(self._command), gcode=gcode)

	#~ capability reporting

	def firmware_capability_handler(self, comm_instance, capability, enabled, already_defined, *args, **kwargs):
		if capability == self.CAP_PROMPT_SUPPORT and enabled:
			self._cap_prompt_support = True
			if self._enable == "detected" and self._enable_signal_support:
				self._printer.commands(["{command} P1".format(command=self._command)])

	#~ prompt handling

	def _show_prompt(self):
		if self._enable == "never" or (self._enable == "detected" and not self._cap_prompt_support):
			return

		self._prompt.activate()
		self._plugin_manager.send_plugin_message(self._identifier, dict(action="show",
		                                                                text=self._prompt.text,
		                                                                choices=self._prompt.choices))

	def _close_prompt(self):
		if self._enable == "never" or (self._enable == "detected" and not self._cap_prompt_support):
			return

		self._prompt = None
		self._plugin_manager.send_plugin_message(self._identifier, dict(action="close"))

	def _answer_prompt(self, choice):
		if self._enable == "never" or (self._enable == "detected" and not self._cap_prompt_support):
			return

		self._close_prompt()
		if "{choice}" in self._command:
			self._printer.commands([self._command.format(choice=choice)])
		else:
		    self._printer.commands(["{command} S{choice}".format(command=self._command,
																 choice=choice)])


__plugin_name__ = "Action Command Prompt Support"
__plugin_description__ = "Allows your printer to trigger prompts via action commands on the connection"
__plugin_author__ = "Gina Häußge"
__plugin_disabling_discouraged__ = gettext("Without this plugin your printer will no longer be able to trigger"
                                           " confirmation or selection prompts in OctoPrint")
__plugin_license__ = "AGPLv3"
__plugin_implementation__ = ActionCommandPromptPlugin()
__plugin_hooks__ = {
	"octoprint.comm.protocol.action": __plugin_implementation__.action_command_handler,
	"octoprint.comm.protocol.gcode.queuing": __plugin_implementation__.gcode_queuing_handler,
	"octoprint.comm.protocol.firmware.capabilities": __plugin_implementation__.firmware_capability_handler,
	"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from octoprint.vendor.sockjs.tornado.transports import pollingbase


class StreamingTransportBase(pollingbase.PollingTransportBase):
    def initialize(self, server):
        super(StreamingTransportBase, self).initialize(server)

        self.amount_limit = self.server.settings['response_limit']

        # HTTP 1.0 client might send keep-alive
        if hasattr(self.request, 'connection') and not self.request.supports_http_1_1():
            self.request.connection.no_keep_alive = True

    def notify_sent(self, data_len):
        """
            Update amount of data sent
        """
        self.amount_limit -= data_len

    def should_finish(self):
        """
            Check if transport should close long running connection after
            sending X bytes to the client.

            `data_len`
                Amount of data that was sent
        """
        if self.amount_limit <= 0:
            return True

        return False

    def send_complete(self):
        """
            Verify if connection should be closed based on amount of data that was sent.
        """
        self.active = True

        if self.should_finish():
            self._detach()

            # Avoid race condition when waiting for write callback and session getting closed in between
            if not self._finished:
                self.safe_finish()
        else:
            if self.session:
                self.session.flush()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import requests
import logging

BRANCH_HEAD_URL = "https://api.github.com/repos/{user}/{repo}/git/refs/heads/{branch}"

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.github_commit")

def _get_latest_commit(user, repo, branch):
	from ..exceptions import NetworkError

	try:
		r = requests.get(BRANCH_HEAD_URL.format(user=user, repo=repo, branch=branch), timeout=(3.05, 30))
	except requests.ConnectionError as exc:
		raise NetworkError(cause=exc)

	from . import log_github_ratelimit
	log_github_ratelimit(logger, r)

	if not r.status_code == requests.codes.ok:
		return None

	reference = r.json()
	if not "object" in reference or not "sha" in reference["object"]:
		return None

	return reference["object"]["sha"]


def get_latest(target, check, online=True):
	from ..exceptions import ConfigurationInvalid

	user = check.get("user")
	repo = check.get("repo")

	if user is None or repo is None:
		raise ConfigurationInvalid("Update configuration for {} of type github_commit needs user and repo set and not None".format(target))

	branch = "master"
	if "branch" in check and check["branch"] is not None:
		branch = check["branch"]

	current = check.get("current")

	information = dict(local=dict(name="Commit {commit}".format(commit=current if current is not None else "?"), value=current),
	                   remote=dict(name="?", value="?"),
	                   needs_online=not check.get("offline", False))
	if not online and information["needs_online"]:
		return information, True

	remote_commit = _get_latest_commit(check["user"], check["repo"], branch)
	remote_name = "Commit {commit}".format(commit=remote_commit) if remote_commit is not None else "-"

	information["remote"] = dict(name=remote_name, value=remote_commit)
	is_current = (current is not None and current == remote_commit) or remote_commit is None

	logger.debug("Target: %s, local: %s, remote: %s" % (target, current, remote_commit))

	return information, is_current




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
click.disable_unicode_literals_warning = True
import logging
import sys

from octoprint.cli import bulk_options, standard_options, set_ctx_obj_option, get_ctx_obj_option

def run_server(basedir, configfile, host, port, v6_only, debug, allow_root, logging_config, verbosity, safe_mode,
               ignore_blacklist, octoprint_daemon=None):
	"""Initializes the environment and starts up the server."""

	from octoprint import init_platform, __display_version__, FatalStartupError

	def log_startup(recorder=None, safe_mode=None, **kwargs):
		from octoprint.logging import get_divider_line

		logger = logging.getLogger("octoprint.startup")

		logger.info(get_divider_line("*"))
		logger.info("Starting OctoPrint {}".format(__display_version__))
		if safe_mode:
			logger.info("Starting in SAFE MODE. Third party plugins will be disabled!")
			if safe_mode == "flag":
				reason = "command line flag"
			elif safe_mode == "settings":
				reason = "setting in config.yaml"
			elif safe_mode == "incomplete_startup":
				reason = "problem during last startup"
			else:
				reason = "unknown"
			logger.info("Reason for safe mode: {}".format(reason))

		if recorder and len(recorder):
			logger.info(get_divider_line("-", "Logged during platform initialization:"))

			from octoprint.logging.handlers import CombinedLogHandler
			handler = CombinedLogHandler(*logging.getLogger().handlers)
			recorder.setTarget(handler)
			recorder.flush()

			logger.info(get_divider_line("-"))

		from octoprint import urllib3_ssl
		if not urllib3_ssl:
			logging.getLogger("octoprint.server")\
				.warn("requests/urllib3 will run in an insecure SSL environment. "
			          "You might see corresponding warnings logged later "
			          "(\"InsecurePlatformWarning\"). It is recommended to either "
			          "update to a Python version >= 2.7.9 or alternatively "
			          "install PyOpenSSL plus its dependencies. For details see "
			          "https://urllib3.readthedocs.org/en/latest/security.html#openssl-pyopenssl")
		logger.info(get_divider_line("*"))

	def log_register_rollover(safe_mode=None, plugin_manager=None, environment_detector=None, **kwargs):
		from octoprint.logging import get_handler, log_to_handler, get_divider_line
		from octoprint.logging.handlers import OctoPrintLogHandler

		def rollover_callback():
			handler = get_handler("file")
			if handler is None:
				return

			logger = logging.getLogger("octoprint.server")

			def _log(message, level=logging.INFO):
				log_to_handler(logger, handler, level, message)

			_log(get_divider_line("-", "Log roll over detected"))
			_log("OctoPrint {}".format(__display_version__))
			if safe_mode:
				_log("SAFE MODE is active. Third party plugins are disabled!")
			plugin_manager.log_all_plugins(only_to_handler=handler)
			environment_detector.log_detected_environment(only_to_handler=handler)
			_log(get_divider_line("-"))

		OctoPrintLogHandler.registerRolloverCallback(rollover_callback)

	try:
		components = init_platform(basedir, configfile,
		                           logging_file=logging_config,
		                           debug=debug,
		                           verbosity=verbosity,
		                           uncaught_logger=__name__,
		                           safe_mode=safe_mode,
                                   ignore_blacklist=ignore_blacklist,
		                           after_safe_mode=log_startup,
		                           after_environment_detector=log_register_rollover)
		settings, _, safe_mode, event_manager, connectivity_checker, plugin_manager, environment_detector = components

	except FatalStartupError as e:
		logger = logging.getLogger("octoprint.startup").fatal
		echo = lambda x: click.echo(x, err=True)

		for method in logger, echo:
			method(str(e))
			method("There was a fatal error starting up OctoPrint.")

	else:
		from octoprint.server import Server
		octoprint_server = Server(settings=settings,
		                          plugin_manager=plugin_manager,
		                          event_manager=event_manager,
		                          connectivity_checker=connectivity_checker,
		                          environment_detector=environment_detector,
		                          host=host,
		                          port=port,
		                          v6_only=v6_only,
		                          debug=debug,
		                          safe_mode=safe_mode,
		                          allow_root=allow_root,
		                          octoprint_daemon=octoprint_daemon)
		octoprint_server.run()

#~~ server options

server_options = bulk_options([
	click.option("--host", type=click.STRING, callback=set_ctx_obj_option,
	             help="Specify the host address on which to bind the server."),
	click.option("--port", type=click.INT, callback=set_ctx_obj_option,
	             help="Specify the port on which to bind the server."),
	click.option("-4", "--ipv4", "v4", is_flag=True, callback=set_ctx_obj_option,
	             help="Bind to IPv4 addresses only. Implies '--host 0.0.0.0'. Silently ignored if -6 is present."),
	click.option("-6", "--ipv6", "v6", is_flag=True, callback=set_ctx_obj_option,
	             help="Bind to IPv6 addresses only. Disables dual stack when binding to any v6 addresses. Silently ignored if -4 is present."),
	click.option("--logging", type=click.Path(), callback=set_ctx_obj_option,
	             help="Specify the config file to use for configuring logging."),
	click.option("--iknowwhatimdoing", "allow_root", is_flag=True, callback=set_ctx_obj_option,
	             help="Allow OctoPrint to run as user root."),
	click.option("--debug", is_flag=True, callback=set_ctx_obj_option,
	             help="Enable debug mode."),
	click.option("--ignore-blacklist", "ignore_blacklist", is_flag=True, callback=set_ctx_obj_option,
	             help="Disable processing of the plugin blacklist.")
])
"""Decorator to add the options shared among the server commands: ``--host``, ``--port``, ``-4``, ``-6``
   ``--logging``, ``--iknowwhatimdoing`` and ``--debug``."""

daemon_options = bulk_options([
	click.option("--pid", type=click.Path(), default="/tmp/octoprint.pid", callback=set_ctx_obj_option,
	             help="Pidfile to use for daemonizing.")
])
"""Decorator to add the options for the daemon subcommand: ``--pid``."""

#~~ "octoprint serve" and "octoprint daemon" commands

@click.group()
def server_commands():
	pass


@server_commands.command(name="safemode")
@standard_options()
@click.pass_context
def enable_safemode(ctx, **kwargs):
	"""Sets the safe mode flag for the next start."""
	from octoprint import init_settings, FatalStartupError

	logging.basicConfig(level=logging.DEBUG if get_ctx_obj_option(ctx, "verbosity", 0) > 0 else logging.WARN)
	try:
		settings = init_settings(get_ctx_obj_option(ctx, "basedir", None), get_ctx_obj_option(ctx, "configfile", None))
	except FatalStartupError as e:
		click.echo(e.message, err=True)
		click.echo("There was a fatal error initializing the settings manager.", err=True)
		ctx.exit(-1)
	else:
		settings.setBoolean(["server", "startOnceInSafeMode"], True)
		settings.save()

		click.echo("Safe mode flag set, OctoPrint will start in safe mode on next restart.")


@server_commands.command(name="serve")
@standard_options()
@server_options
@click.pass_context
def serve_command(ctx, **kwargs):
	"""Starts the OctoPrint server."""

	def get_value(key):
		return get_ctx_obj_option(ctx, key, kwargs.get(key))

	host = get_value("host")
	port = get_value("port")
	v4 = get_value("v4")
	v6 = get_value("v6")
	logging = get_value("logging")
	allow_root = get_value("allow_root")
	debug = get_value("debug")

	basedir = get_value("basedir")
	configfile = get_value("configfile")
	verbosity = get_value("verbosity")
	safe_mode = "flag" if get_value("safe_mode") else None
	ignore_blacklist = get_value("ignore_blacklist")

	if v4 and not host:
		host = "0.0.0.0"

	run_server(basedir, configfile, host, port, v6, debug,
	           allow_root, logging, verbosity, safe_mode,
	           ignore_blacklist)


if sys.platform != "win32" and sys.platform != "darwin":
	# we do not support daemon mode under windows or macosx

	@server_commands.command(name="daemon")
	@standard_options()
	@server_options
	@daemon_options
	@click.argument("command", type=click.Choice(["start", "stop", "restart", "status"]),
	                metavar="start|stop|restart|status")
	@click.pass_context
	def daemon_command(ctx, command, **kwargs):
		"""
		Starts, stops or restarts in daemon mode.

		Please note that daemon mode is not supported under Windows and MacOSX right now.
		"""

		def get_value(key):
			return get_ctx_obj_option(ctx, key, kwargs.get(key))

		host = get_value("host")
		port = get_value("port")
		v4 = get_value("v4")
		v6 = get_value("v6")
		logging = get_value("logging")
		allow_root = get_value("allow_root")
		debug = get_value("debug")
		pid = get_value("pid")

		basedir = get_value("basedir")
		configfile = get_value("configfile")
		verbosity = get_value("verbosity")
		safe_mode = "flag" if get_value("safe_mode") else None
		ignore_blacklist = get_value("ignore_blacklist")

		if v4 and not host:
			host = "0.0.0.0"

		if pid is None:
			click.echo("No path to a pidfile set",
			           file=sys.stderr)
			sys.exit(1)

		from octoprint.daemon import Daemon
		class OctoPrintDaemon(Daemon):
			def __init__(self, pidfile, basedir, configfile, host, port, v6_only, debug, allow_root, logging_config,
			             verbosity, safe_mode, ignore_blacklist):
				Daemon.__init__(self, pidfile)

				self._basedir = basedir
				self._configfile = configfile
				self._host = host
				self._port = port
				self._v6_only = v6_only
				self._debug = debug
				self._allow_root = allow_root
				self._logging_config = logging_config
				self._verbosity = verbosity
				self._safe_mode = safe_mode
				self._ignore_blacklist = ignore_blacklist

			def run(self):
				run_server(self._basedir, self._configfile, self._host, self._port, self._v6_only, self._debug,
				           self._allow_root, self._logging_config, self._verbosity, self._safe_mode,
				           self._ignore_blacklist, octoprint_daemon=self)

		octoprint_daemon = OctoPrintDaemon(pid, basedir, configfile, host, port, v6, debug, allow_root, logging,
		                                   verbosity, safe_mode, ignore_blacklist)

		if command == "start":
			octoprint_daemon.start()
		elif command == "stop":
			octoprint_daemon.stop()
		elif command == "restart":
			octoprint_daemon.restart()
		elif command == "status":
			octoprint_daemon.status()


